from __future__ import annotations

from .types import ReloadMode

# control defaults
DEFAULT_ENV_PREFIX = ""
DEFAULT_CASE_SENSITIVE = False
DEFAULT_RELOAD_MODE: ReloadMode = "off"

# env/runtime naming
ENV_KEY_SEPARATOR = "__"
DOTENV_FILENAME = ".env"
DOTENV_EXPORT_PREFIX = "export "
SETTINGS_FILENAME = "settings.yaml"

# scalar tokens
TRUE_TEXT_VALUES = frozenset({"1", "true", "yes", "on"})
FALSE_TEXT_VALUES = frozenset({"0", "false", "no", "off"})
NULL_TEXT_VALUES = frozenset({"null", "none"})

# reload-mode normalization
RELOAD_MODE_ALWAYS_TOKENS = frozenset({"always"})
RELOAD_MODE_ON_CHANGE_TOKENS = frozenset({"on_change", "on-change", "onchange", "true", "1", "yes"})
RELOAD_MODE_OFF_TOKENS = frozenset({"off", "false", "0", "no"})

__all__ = [
    "DEFAULT_ENV_PREFIX",
    "DEFAULT_CASE_SENSITIVE",
    "DEFAULT_RELOAD_MODE",
    "DOTENV_EXPORT_PREFIX",
    "DOTENV_FILENAME",
    "ENV_KEY_SEPARATOR",
    "FALSE_TEXT_VALUES",
    "NULL_TEXT_VALUES",
    "RELOAD_MODE_ALWAYS_TOKENS",
    "RELOAD_MODE_OFF_TOKENS",
    "RELOAD_MODE_ON_CHANGE_TOKENS",
    "SETTINGS_FILENAME",
    "TRUE_TEXT_VALUES",
]



from . import exceptions
from .access import GetSettings, GetSettingsMap, SettingsRef
from .base import BaseSettings
from .bootstrap import init_settings, reload_settings
from .registry import Settings, SettingsMap

__all__ = [
    "BaseSettings",
    "GetSettings",
    "GetSettingsMap",
    "SettingsRef",
    "Settings",
    "SettingsMap",
    "exceptions",
    "init_settings",
    "reload_settings",
]



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

from .context import ConfigContext
from .types import ProjectionKind, SourceDescriptor, SourceName, SourceToken


@dataclass(frozen=True)
class SourcePolicy:
    auto_refresh: bool
    manual_refresh: bool
    follow_context: bool
    participates_in_controls: bool = True


@dataclass(frozen=True)
class SourceBinding:
    source: SourceName
    descriptor: SourceDescriptor


@dataclass(frozen=True)
class LoadedSource:
    token: SourceToken
    payload: dict[str, Any]


BindingBuilder = Callable[[ConfigContext], SourceBinding]
SourceProbe = Callable[[SourceBinding], SourceToken]
SourceLoader = Callable[[SourceBinding], LoadedSource]
BindingValidator = Callable[[ConfigContext, SourceBinding], None]


@dataclass(frozen=True)
class SourceSpec:
    name: SourceName
    priority: int
    projection_kind: ProjectionKind
    policy: SourcePolicy
    bind: BindingBuilder
    probe: SourceProbe
    load: SourceLoader
    validate_final_binding: BindingValidator | None = None


class SourceRegistry:
    def __init__(self) -> None:
        self._specs: dict[SourceName, SourceSpec] = {}
        self._version = 0

    def register(self, spec: SourceSpec) -> None:
        current = self._specs.get(spec.name)
        if current == spec:
            return
        self._specs[spec.name] = spec
        self._version += 1

    def unregister(self, name: SourceName) -> None:
        if name not in self._specs:
            return
        del self._specs[name]
        self._version += 1

    def get(self, name: SourceName) -> SourceSpec | None:
        return self._specs.get(name)

    def ordered(self) -> tuple[SourceSpec, ...]:
        return tuple(sorted(self._specs.values(), key=lambda spec: (spec.priority, spec.name)))

    def version(self) -> int:
        return self._version


__all__ = [
    "BindingBuilder",
    "BindingValidator",
    "LoadedSource",
    "SourceBinding",
    "SourceLoader",
    "SourcePolicy",
    "SourceProbe",
    "SourceRegistry",
    "SourceSpec",
]



from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass
from typing import Any

from pydantic import BaseModel

from .context import ConfigContext
from .source_contract import SourceBinding
from .types import SourceName, SourceToken


@dataclass(frozen=True)
class SourceSnapshot:
    source: SourceName
    rev: int
    binding: SourceBinding
    token: SourceToken
    payload: dict[str, Any]


@dataclass(frozen=True)
class RuntimeState:
    sources_version: int
    last_rev: int
    context: ConfigContext
    snapshots: dict[SourceName, SourceSnapshot]
    settings: BaseModel


@dataclass(frozen=True)
class SnapshotInspection:
    source: SourceName
    rev: int
    binding: SourceBinding
    token: SourceToken
    payload: dict[str, Any]


@dataclass(frozen=True)
class RuntimeInspection:
    sources_version: int
    last_rev: int
    context: ConfigContext
    snapshots: tuple[SnapshotInspection, ...]


def inspect_runtime_state(runtime: RuntimeState | None) -> RuntimeInspection | None:
    if runtime is None:
        return None

    rows = tuple(
        SnapshotInspection(
            source=snapshot.source,
            rev=snapshot.rev,
            binding=snapshot.binding,
            token=deepcopy(snapshot.token),
            payload=deepcopy(snapshot.payload),
        )
        for snapshot in sorted(runtime.snapshots.values(), key=lambda item: item.source)
    )
    return RuntimeInspection(
        sources_version=runtime.sources_version,
        last_rev=runtime.last_rev,
        context=runtime.context,
        snapshots=rows,
    )


__all__ = [
    "ConfigContext",
    "inspect_runtime_state",
    "RuntimeInspection",
    "RuntimeState",
    "SnapshotInspection",
    "SourceSnapshot",
]



from __future__ import annotations

from dataclasses import replace
import logging
import os
import sys
import types
from pathlib import Path

import pytest
from pydantic import Field

from fastapiex.settings import (
    BaseSettings,
    GetSettings,
    GetSettingsMap,
    Settings,
    SettingsMap,
    exceptions,
    init_settings,
    reload_settings,
)
from fastapiex.settings import manager as manager_module
from fastapiex.settings import registry as registry_module
from fastapiex.settings import control_contract as control_contract_module
from fastapiex.settings.manager import get_settings_manager
from fastapiex.settings.source_contract import LoadedSource, SourceBinding, SourcePolicy, SourceSpec

SettingsRegistrationError = exceptions.SettingsRegistrationError
SettingsResolveError = exceptions.SettingsResolveError

DYNAMIC_MODULE = "tests.dynamic_settings_runtime"


@pytest.fixture(autouse=True)
def _reset_runtime(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(manager_module, "_GLOBAL_MANAGER", manager_module.SettingsManager())
    monkeypatch.setattr(registry_module, "_GLOBAL_REGISTRY", registry_module.SettingsRegistry())
    sys.modules.pop(DYNAMIC_MODULE, None)

    for key in list(os.environ):
        if key.startswith("FASTAPIEX__") or key.startswith("TEST__"):
            monkeypatch.delenv(key, raising=False)
    monkeypatch.setenv("FASTAPIEX__SETTINGS__ENV_PREFIX", "TEST__")


def test_nested_sections_are_composed(tmp_path: Path) -> None:
    @Settings("father")
    class FatherSettings(BaseSettings):
        a: int = Field(default=1, ge=1)

    @Settings("father.son")
    class SonSettings(BaseSettings):
        a: int = Field(default=2, ge=1)

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("father:\n  a: 7\n  son:\n    a: 9\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    assert GetSettings(target=FatherSettings, field="a") == 7
    assert GetSettings(target=SonSettings, field="a") == 9
    assert GetSettings(target="father.son", field="a") == 9


def test_settingsmap_and_settings_read_same_map_section(tmp_path: Path) -> None:
    @SettingsMap("services")
    class ServiceSettings(BaseSettings):
        host: str
        port: int

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text(
        "services:\n  api:\n    host: 127.0.0.1\n    port: 8000\n  admin:\n    host: 127.0.0.2\n    port: 9000\n",
        encoding="utf-8",
    )
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    services = GetSettingsMap(target=ServiceSettings)
    assert services["api"].host == "127.0.0.1"

    assert GetSettings(target="services.api", field="host") == "127.0.0.1"
    assert GetSettings(target=ServiceSettings, field="api.host") == "127.0.0.1"
    same_services = GetSettings(target=ServiceSettings)
    assert isinstance(same_services, dict)
    assert same_services["admin"].host == "127.0.0.2"


def test_settings_singleton_map_resolves_to_mapping_without_special_unwrap(tmp_path: Path) -> None:
    @SettingsMap("services")
    class ServiceSettings(BaseSettings):
        host: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("services:\n  api:\n    host: localhost\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    services = GetSettings(target=ServiceSettings)
    assert isinstance(services, dict)
    assert services["api"].host == "localhost"

    with pytest.raises(SettingsResolveError):
        GetSettings(target=ServiceSettings, field="host")


def test_target_none_with_default_returns_default_as_is_even_with_field() -> None:
    assert GetSettings(target=None, field="x.y", default={"x": {"y": 3}}) == {"x": {"y": 3}}
    assert GetSettingsMap(target=None, default={"k": 1})["k"] == 1


def test_target_none_without_default_raises() -> None:
    with pytest.raises(SettingsResolveError):
        GetSettings(target=None, field="x.y")


def test_declaration_decorators_reject_invalid_arguments() -> None:
    class Plain:
        pass

    with pytest.raises(TypeError):
        Settings(target="app")  # type: ignore[call-arg]

    with pytest.raises(TypeError):
        SettingsMap(target="app")  # type: ignore[call-arg]

    with pytest.raises(TypeError, match="dotted path string, model class, or no argument"):
        Settings(123)  # type: ignore[arg-type]

    with pytest.raises(TypeError, match="BaseModel subclasses"):
        Settings(Plain)  # type: ignore[arg-type]

    with pytest.raises(TypeError, match="dotted path string, model class, or no argument"):
        SettingsMap(123)  # type: ignore[arg-type]

    with pytest.raises(TypeError, match="BaseModel subclasses"):
        SettingsMap(Plain)  # type: ignore[arg-type]


def test_basesettings_subclass_without_decorator_is_not_registered(tmp_path: Path) -> None:
    class UndeclaredSettings(BaseSettings):
        value: int = 1

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("{}\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    with pytest.raises(SettingsResolveError, match="did not match any declared section"):
        GetSettings(target=UndeclaredSettings)

    assert GetSettings(target=UndeclaredSettings, default="fallback") == "fallback"


def test_validation_error_falls_back_to_default_and_logs_warning(
    caplog: pytest.LogCaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("app: {}\n", encoding="utf-8")
    monkeypatch.setenv("FASTAPIEX__SETTINGS__PATH", str(settings_file))

    caplog.set_level(logging.WARNING)

    assert GetSettings(target="app", field="name", default="fallback") == "fallback"
    assert GetSettings(target="app", field="name", default="fallback") == "fallback"

    warning_messages = [record.getMessage() for record in caplog.records if record.levelno >= logging.WARNING]
    validation_warnings = [msg for msg in warning_messages if "validation failed" in msg]
    assert len(validation_warnings) == 1


def test_registration_allows_case_variant_section_names(tmp_path: Path) -> None:
    @Settings("APP")
    class UpperApp(BaseSettings):
        name: str

    @Settings("app")
    class LowerApp(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("APP:\n  name: upper\napp:\n  name: lower\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    # Type-target reads remain exact and do not become ambiguous in case-insensitive mode.
    assert GetSettings(target=UpperApp, field="name") == "upper"
    assert GetSettings(target=LowerApp, field="name") == "lower"


def test_case_insensitive_path_lookup_is_ambiguous_for_case_variant_sections(tmp_path: Path) -> None:
    @Settings("APP")
    class UpperApp(BaseSettings):
        name: str

    @Settings("app")
    class LowerApp(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("APP:\n  name: upper\napp:\n  name: lower\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    with pytest.raises(SettingsResolveError):
        GetSettings(target="app", field="name")


def test_case_insensitive_map_lookup_is_ambiguous_for_case_variant_keys(tmp_path: Path) -> None:
    @SettingsMap("services")
    class ServiceSettings(BaseSettings):
        host: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text(
        "services:\n  API:\n    host: upper\n  api:\n    host: lower\n",
        encoding="utf-8",
    )
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    with pytest.raises(SettingsResolveError):
        GetSettings(target="services.api", field="host")


def test_case_insensitive_env_override_applies_to_uppercase_declaration(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("TEST__APP__NAME", "env-value")

    @Settings("APP")
    class AppSettings(BaseSettings):
        NAME: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("APP:\n  NAME: yaml\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    assert GetSettings(target=AppSettings, field="NAME") == "env-value"
    assert GetSettings(target="app", field="name") == "env-value"


def test_prefixed_settings_path_env_is_used_for_implicit_init(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("app:\n  name: from-env-path\n", encoding="utf-8")
    monkeypatch.setenv("FASTAPIEX__SETTINGS__PATH", str(settings_file))

    assert GetSettings(target=AppSettings, field="name") == "from-env-path"


def test_implicit_init_applies_runtime_controls_from_snapshot(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    @Settings("App")
    class AppSettings(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text(
        "fastapiex:\n  settings:\n    case_sensitive: true\n    reload: always\nApp:\n  name: v1\n",
        encoding="utf-8",
    )
    monkeypatch.setenv("FASTAPIEX__SETTINGS__PATH", str(settings_file))

    assert GetSettings(target=AppSettings, field="name") == "v1"

    settings_file.write_text(
        "fastapiex:\n  settings:\n    case_sensitive: true\n    reload: always\nApp:\n  name: v2\n",
        encoding="utf-8",
    )
    assert GetSettings(target=AppSettings, field="name") == "v2"


def test_prefixed_base_dir_env_is_used_for_implicit_init(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("app:\n  name: from-base-dir\n", encoding="utf-8")
    monkeypatch.setenv("FASTAPIEX__BASE_DIR", str(tmp_path))

    assert GetSettings(target=AppSettings, field="name") == "from-base-dir"


def test_snapshot_control_env_prefix_reprojects_env_from_raw(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.delenv("FASTAPIEX__SETTINGS__ENV_PREFIX", raising=False)
    monkeypatch.setenv("TEST__APP__NAME", "from-test")
    monkeypatch.setenv("ALT__APP__NAME", "from-alt")

    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text(
        "fastapiex:\n  settings:\n    env_prefix: ALT__\napp:\n  name: from-yaml\n",
        encoding="utf-8",
    )
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    manager = get_settings_manager()
    runtime_view = manager.inspect_runtime()
    assert runtime_view is not None
    assert runtime_view.context.env_prefix == "ALT__"
    assert GetSettings(target=AppSettings, field="name") == "from-alt"


def test_case_sensitive_true_on_posix_allows_distinct_section_names(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("FASTAPIEX__SETTINGS__CASE_SENSITIVE", "true")
    monkeypatch.setattr(os, "name", "posix", raising=False)

    @Settings("APP")
    class UpperApp(BaseSettings):
        name: str

    @Settings("app")
    class LowerApp(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("APP:\n  name: upper\napp:\n  name: lower\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    assert GetSettings(target="APP", field="name") == "upper"
    assert GetSettings(target="app", field="name") == "lower"
    assert GetSettings(target=UpperApp, field="name") == "upper"
    assert GetSettings(target=LowerApp, field="name") == "lower"


def test_snapshot_control_case_sensitive_is_applied_before_env_projection(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setattr(os, "name", "posix", raising=False)
    monkeypatch.setenv("TEST__APP__name", "upper")
    monkeypatch.setenv("TEST__app__name", "lower")

    @Settings("APP")
    class UpperApp(BaseSettings):
        name: str

    @Settings("app")
    class LowerApp(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("fastapiex:\n  settings:\n    case_sensitive: true\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    assert GetSettings(target=UpperApp, field="name") == "upper"
    assert GetSettings(target=LowerApp, field="name") == "lower"
    assert GetSettings(target="FASTAPIEX.SETTINGS.CASE_SENSITIVE") is True


def test_case_sensitive_true_is_ignored_on_windows(
    caplog: pytest.LogCaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("FASTAPIEX__SETTINGS__CASE_SENSITIVE", "true")
    monkeypatch.setattr(control_contract_module, "os", types.SimpleNamespace(name="nt"))

    caplog.set_level(logging.WARNING)

    @Settings("APP")
    class UpperApp(BaseSettings):
        name: str

    @Settings("app")
    class LowerApp(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("APP:\n  name: upper\napp:\n  name: lower\n", encoding="utf-8")
    monkeypatch.chdir(tmp_path)
    init_settings()

    assert GetSettings(target=UpperApp, field="name") == "upper"
    assert GetSettings(target=LowerApp, field="name") == "lower"
    with pytest.raises(SettingsResolveError):
        GetSettings(target="app", field="name")

    warning_messages = [record.getMessage() for record in caplog.records if record.levelno >= logging.WARNING]
    assert any("ignored on Windows" in message for message in warning_messages)


def test_reserved_prefix_is_rejected() -> None:
    with pytest.raises(SettingsRegistrationError):

        @Settings("FASTAPIEX.internal")
        class InternalSettings(BaseSettings):
            enabled: bool = True


def test_invalid_registration_does_not_poison_registry(tmp_path: Path) -> None:
    with pytest.raises(SettingsRegistrationError):

        @Settings("FASTAPIEX.internal")
        class InvalidSettings(BaseSettings):
            enabled: bool = True

    @Settings("app")
    class AppSettings(BaseSettings):
        name: str = "demo"

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("app:\n  name: ok\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()
    assert GetSettings(target=AppSettings, field="name") == "ok"


def test_settings_prefix_is_not_reserved_anymore() -> None:
    @Settings("settings.runtime")
    class RuntimeSettings(BaseSettings):
        enabled: bool = True

    assert GetSettings(target=RuntimeSettings, field="enabled", default=False) is True


def test_manual_reload_forces_source_reread(tmp_path: Path) -> None:
    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("app:\n  name: v1\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    assert GetSettings(target=AppSettings, field="name") == "v1"

    settings_file.write_text("app:\n  name: v2\n", encoding="utf-8")
    assert GetSettings(target=AppSettings, field="name") == "v1"

    reload_settings(reason="test")
    assert GetSettings(target=AppSettings, field="name") == "v2"


def test_snapshot_fastapiex_settings_path_can_switch_source_after_explicit_bootstrap(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    bootstrap = tmp_path / "bootstrap.yaml"
    env_file = tmp_path / "env.yaml"
    switched = tmp_path / "switched"
    switched.mkdir()
    switched_yaml = switched / "settings.yaml"

    env_file.write_text("app:\n  name: from-env\n", encoding="utf-8")
    switched_yaml.write_text(
        f'app:\n  name: from-switched\nfastapiex:\n  settings:\n    path: "{switched}"\n',
        encoding="utf-8",
    )
    bootstrap.write_text(
        f'app:\n  name: from-bootstrap\nfastapiex:\n  settings:\n    path: "{switched}"\n',
        encoding="utf-8",
    )

    monkeypatch.setenv("FASTAPIEX__SETTINGS__PATH", str(env_file))
    init_settings(settings_path=bootstrap)

    assert GetSettings(target=AppSettings, field="name") == "from-bootstrap"


def test_snapshot_fastapiex_settings_path_drives_runtime_source_switch(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("FASTAPIEX__SETTINGS__RELOAD", "on_change")

    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    first = tmp_path / "first.yaml"
    second = tmp_path / "second.yaml"

    first.write_text(
        f'app:\n  name: first\nfastapiex:\n  settings:\n    path: "{first}"\n',
        encoding="utf-8",
    )
    second.write_text(
        f'app:\n  name: second\nfastapiex:\n  settings:\n    path: "{second}"\n',
        encoding="utf-8",
    )
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(first)
    init_settings()

    assert GetSettings(target=AppSettings, field="name") == "first"

    first.write_text(
        f'app:\n  name: first-updated\nfastapiex:\n  settings:\n    path: "{second}"\n',
        encoding="utf-8",
    )

    assert GetSettings(target=AppSettings, field="name") == "second"


def test_runtime_reload_mode_from_snapshot_is_not_affected_by_env_flip_after_init(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("FASTAPIEX__SETTINGS__RELOAD", "on_change")

    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("app:\n  name: v1\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()
    assert GetSettings(target=AppSettings, field="name") == "v1"

    settings_file.write_text("app:\n  name: v2\n", encoding="utf-8")
    assert GetSettings(target=AppSettings, field="name") == "v2"

    monkeypatch.setenv("FASTAPIEX__SETTINGS__RELOAD", "off")
    settings_file.write_text("app:\n  name: v3\n", encoding="utf-8")
    assert GetSettings(target=AppSettings, field="name") == "v3"

    settings_file.write_text("app:\n  name: v4\n", encoding="utf-8")
    assert GetSettings(target=AppSettings, field="name") == "v4"

    reload_settings(reason="apply-runtime-options")

    settings_file.write_text("app:\n  name: v5\n", encoding="utf-8")
    assert GetSettings(target=AppSettings, field="name") == "v5"


def test_reload_mode_change_in_snapshot_takes_effect_without_manual_reload(tmp_path: Path) -> None:
    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("fastapiex:\n  settings:\n    reload: always\napp:\n  name: v1\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()
    assert GetSettings(target=AppSettings, field="name") == "v1"

    settings_file.write_text("fastapiex:\n  settings:\n    reload: off\napp:\n  name: v2\n", encoding="utf-8")
    assert GetSettings(target=AppSettings, field="name") == "v2"

    settings_file.write_text("fastapiex:\n  settings:\n    reload: off\napp:\n  name: v3\n", encoding="utf-8")
    assert GetSettings(target=AppSettings, field="name") == "v2"


def test_reload_mode_off_in_snapshot_is_not_enabled_by_env_flip(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("FASTAPIEX__SETTINGS__RELOAD", "off")

    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("app:\n  name: v1\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()
    assert GetSettings(target=AppSettings, field="name") == "v1"

    monkeypatch.setenv("FASTAPIEX__SETTINGS__RELOAD", "on_change")
    settings_file.write_text("app:\n  name: v2\n", encoding="utf-8")
    assert GetSettings(target=AppSettings, field="name") == "v1"

    reload_settings(reason="enable-auto")
    assert GetSettings(target=AppSettings, field="name") == "v2"

    settings_file.write_text("app:\n  name: v3\n", encoding="utf-8")
    assert GetSettings(target=AppSettings, field="name") == "v2"


def test_fastapiex_namespace_lookup_is_always_case_insensitive(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("FASTAPIEX__SETTINGS__CASE_SENSITIVE", "true")
    monkeypatch.setenv("FASTAPIEX__SETTINGS__RELOAD", "off")

    @Settings("APP")
    class UpperApp(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("APP:\n  name: upper\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    assert GetSettings(target="FastAPIEx.Settings.Reload") is False
    assert GetSettings(target="FASTAPIEX.SETTINGS.CASE_SENSITIVE") is True
    assert GetSettings(target=UpperApp, field="name") == "upper"


def test_fastapiex_namespace_with_mixed_case_yaml_keys_is_queryable(
    tmp_path: Path,
) -> None:
    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text(
        "FastAPIEx:\n  Settings:\n    Reload: always\napp:\n  name: value\n",
        encoding="utf-8",
    )
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    assert GetSettings(target="FASTAPIEX.SETTINGS.RELOAD") == "always"
    assert GetSettings(target=AppSettings, field="name") == "value"


def test_fastapiex_namespace_with_mixed_case_yaml_keys_stays_queryable_in_case_sensitive_mode(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("FASTAPIEX__SETTINGS__CASE_SENSITIVE", "true")

    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text(
        "FastAPIEx:\n  Settings:\n    Reload: always\napp:\n  name: value\n",
        encoding="utf-8",
    )
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    assert GetSettings(target="FASTAPIEX.SETTINGS.RELOAD") == "always"
    assert GetSettings(target=AppSettings, field="name") == "value"


def test_nested_fastapiex_business_field_is_not_normalized_as_control_namespace(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("FASTAPIEX__SETTINGS__CASE_SENSITIVE", "true")

    @Settings("app")
    class AppSettings(BaseSettings):
        fastapiex: dict[str, str]

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text(
        "app:\n  fastapiex:\n    TokenKey: Value\n",
        encoding="utf-8",
    )
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    payload = GetSettings(target=AppSettings, field="fastapiex")
    assert payload["TokenKey"] == "Value"
    assert "tokenkey" not in payload


def test_lww_yaml_change_overrides_older_env_value(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("FASTAPIEX__SETTINGS__RELOAD", "on_change")
    monkeypatch.setenv("TEST__APP__NAME", "env-v1")

    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("app:\n  name: yaml-v1\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    assert GetSettings(target=AppSettings, field="name") == "env-v1"

    settings_file.write_text("app:\n  name: yaml-v2\n", encoding="utf-8")
    assert GetSettings(target=AppSettings, field="name") == "yaml-v2"


def test_manual_reload_does_not_reingest_env_or_dotenv_sources(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("FASTAPIEX__SETTINGS__RELOAD", "on_change")
    monkeypatch.setenv("TEST__APP__NAME", "env-v1")

    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("app:\n  name: yaml-v1\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()
    assert GetSettings(target=AppSettings, field="name") == "env-v1"

    settings_file.write_text("app:\n  name: yaml-v2\n", encoding="utf-8")
    assert GetSettings(target=AppSettings, field="name") == "yaml-v2"

    monkeypatch.setenv("TEST__APP__NAME", "env-v2")
    reload_settings(reason="rebuild")
    assert GetSettings(target=AppSettings, field="name") == "yaml-v2"


def test_dotenv_changes_are_not_watched_but_yaml_changes_are(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("FASTAPIEX__SETTINGS__RELOAD", "on_change")

    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("app:\n  name: yaml-v1\n", encoding="utf-8")
    dotenv_file = tmp_path / ".env"
    dotenv_file.write_text("TEST__APP__NAME=dotenv-v1\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    assert GetSettings(target=AppSettings, field="name") == "dotenv-v1"

    dotenv_file.write_text("TEST__APP__NAME=dotenv-v2\n", encoding="utf-8")
    assert GetSettings(target=AppSettings, field="name") == "dotenv-v1"

    settings_file.write_text("app:\n  name: yaml-v2\n", encoding="utf-8")
    assert GetSettings(target=AppSettings, field="name") == "yaml-v2"


def test_dotenv_source_can_enable_runtime_sync(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("FASTAPIEX__SETTINGS__RELOAD", "on_change")

    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("app:\n  name: yaml-v1\n", encoding="utf-8")
    dotenv_file = tmp_path / ".env"
    dotenv_file.write_text("TEST__APP__NAME=dotenv-v1\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()
    assert GetSettings(target=AppSettings, field="name") == "dotenv-v1"

    dotenv_file.write_text("TEST__APP__NAME=dotenv-v2\n", encoding="utf-8")
    assert GetSettings(target=AppSettings, field="name") == "dotenv-v1"

    manager = get_settings_manager()
    builtin_dotenv = manager.get_source("dotenv")
    assert builtin_dotenv is not None
    manager.register_source(
        replace(
            builtin_dotenv,
            policy=replace(
                builtin_dotenv.policy,
                auto_refresh=True,
                manual_refresh=True,
            ),
        )
    )

    dotenv_file.write_text("TEST__APP__NAME=dotenv-v3\n", encoding="utf-8")
    assert GetSettings(target=AppSettings, field="name") == "dotenv-v3"


def test_custom_source_spec_can_be_registered() -> None:
    manager = get_settings_manager()
    manager.register_source(
        SourceSpec(
            name="custom",
            priority=4,
            projection_kind="mapping",
            policy=SourcePolicy(
                auto_refresh=False,
                manual_refresh=False,
                follow_context=False,
            ),
            bind=lambda context: SourceBinding(source="custom", descriptor=("custom", context.settings_path)),
            probe=lambda binding: binding.descriptor,
            load=lambda binding: LoadedSource(token=binding.descriptor, payload={}),
        )
    )
    assert manager.get_source("custom") is not None


def test_replacing_source_spec_rebinds_on_forced_full_refresh(tmp_path: Path) -> None:
    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    manager = get_settings_manager()
    manager.register_source(
        SourceSpec(
            name="custom",
            priority=4,
            projection_kind="mapping",
            policy=SourcePolicy(
                auto_refresh=False,
                manual_refresh=True,
                follow_context=False,
            ),
            bind=lambda context: SourceBinding(source="custom", descriptor=("kind-a", "a")),
            probe=lambda binding: binding.descriptor,
            load=lambda binding: LoadedSource(
                token=binding.descriptor,
                payload={"app": {"name": str(binding.descriptor)}},
            ),
        )
    )

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("app:\n  name: yaml\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()
    reload_settings(reason="load-custom-a")
    assert GetSettings(target=AppSettings, field="name") == "('kind-a', 'a')"

    manager.register_source(
        SourceSpec(
            name="custom",
            priority=4,
            projection_kind="mapping",
            policy=SourcePolicy(
                auto_refresh=False,
                manual_refresh=True,
                follow_context=False,
            ),
            bind=lambda context: SourceBinding(source="custom", descriptor=("kind-b", "b")),
            probe=lambda binding: binding.descriptor,
            load=lambda binding: LoadedSource(
                token=binding.descriptor,
                payload={"app": {"name": str(binding.descriptor)}},
            ),
        )
    )
    reload_settings(reason="load-custom-b")

    runtime_view = manager.inspect_runtime()
    assert runtime_view is not None
    assert GetSettings(target=AppSettings, field="name") == "('kind-b', 'b')"
    custom_snapshot = next(snapshot for snapshot in runtime_view.snapshots if snapshot.source == "custom")
    assert custom_snapshot.binding.descriptor == ("kind-b", "b")


def test_unregistered_source_is_pruned_from_runtime_snapshots(tmp_path: Path) -> None:
    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    manager = get_settings_manager()
    manager.register_source(
        SourceSpec(
            name="ephemeral",
            priority=4,
            projection_kind="mapping",
            policy=SourcePolicy(
                auto_refresh=False,
                manual_refresh=True,
                follow_context=False,
            ),
            bind=lambda context: SourceBinding(source="ephemeral", descriptor=("ephemeral", 1)),
            probe=lambda binding: binding.descriptor,
            load=lambda binding: LoadedSource(
                token=binding.descriptor,
                payload={"app": {"name": "ephemeral"}},
            ),
        )
    )

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("app:\n  name: yaml\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()
    reload_settings(reason="load-ephemeral")
    assert GetSettings(target=AppSettings, field="name") == "ephemeral"

    runtime_view = manager.inspect_runtime()
    assert runtime_view is not None
    assert any(snapshot.source == "ephemeral" for snapshot in runtime_view.snapshots)

    manager.unregister_source("ephemeral")
    reload_settings(reason="prune-ephemeral")
    assert GetSettings(target=AppSettings, field="name") == "yaml"

    runtime_view = manager.inspect_runtime()
    assert runtime_view is not None
    assert all(snapshot.source != "ephemeral" for snapshot in runtime_view.snapshots)


def test_custom_dotenv_source_spec_drives_auto_refresh(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("FASTAPIEX__SETTINGS__RELOAD", "on_change")

    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    snapshots = [
        ({"TEST__APP__NAME": "dotenv-v1"}, "v1"),
        ({"TEST__APP__NAME": "dotenv-v2"}, "v2"),
    ]
    state = {"index": 0}

    manager = get_settings_manager()
    builtin_dotenv = manager.get_source("dotenv")
    assert builtin_dotenv is not None

    def _custom_dotenv_probe(_: SourceBinding) -> str:
        return snapshots[state["index"]][1]

    def _custom_dotenv_load(_: SourceBinding) -> LoadedSource:
        payload, token = snapshots[state["index"]]
        return LoadedSource(token=token, payload=dict(payload))

    manager.register_source(
        replace(
            builtin_dotenv,
            policy=replace(
                builtin_dotenv.policy,
                auto_refresh=True,
                manual_refresh=True,
            ),
            probe=_custom_dotenv_probe,
            load=_custom_dotenv_load,
        )
    )

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("app:\n  name: yaml\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    assert GetSettings(target=AppSettings, field="name") == "dotenv-v1"

    state["index"] = 1
    assert GetSettings(target=AppSettings, field="name") == "dotenv-v2"


def test_custom_dotenv_source_spec_can_revert_to_builtin_loader(tmp_path: Path) -> None:
    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    manager = get_settings_manager()
    builtin_dotenv = manager.get_source("dotenv")
    assert builtin_dotenv is not None
    state = {"value": "custom-v1"}
    manager.register_source(
        replace(
            builtin_dotenv,
            policy=replace(
                builtin_dotenv.policy,
                auto_refresh=True,
                manual_refresh=True,
            ),
            probe=lambda binding: state["value"],
            load=lambda binding: LoadedSource(
                token=state["value"],
                payload={"TEST__APP__NAME": state["value"]},
            ),
        )
    )

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("app:\n  name: yaml\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()
    reload_settings(reason="use-custom-dotenv")
    assert GetSettings(target=AppSettings, field="name") == "custom-v1"

    manager.register_source(
        replace(
            builtin_dotenv,
            policy=replace(
                builtin_dotenv.policy,
                auto_refresh=True,
                manual_refresh=True,
            ),
        )
    )
    state["value"] = "custom-v2"
    reload_settings(reason="restore-default-dotenv")
    assert GetSettings(target=AppSettings, field="name") == "yaml"


def test_custom_yaml_loader_in_explicit_file_mode_uses_real_file_existence(tmp_path: Path) -> None:
    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    manager = get_settings_manager()
    builtin_yaml = manager.get_source("yaml")
    assert builtin_yaml is not None
    manager.register_source(
        replace(
            builtin_yaml,
            probe=lambda binding: "token",
            load=lambda binding: LoadedSource(
                token="token",
                payload={"app": {"name": "custom"}},
            ),
        )
    )

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("app:\n  name: yaml\n", encoding="utf-8")

    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()
    assert GetSettings(target=AppSettings, field="name") == "custom"


def test_custom_yaml_loader_can_own_explicit_file_mode_without_local_file(tmp_path: Path) -> None:
    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    manager = get_settings_manager()
    builtin_yaml = manager.get_source("yaml")
    assert builtin_yaml is not None
    manager.register_source(
        replace(
            builtin_yaml,
            bind=lambda context: SourceBinding(source="yaml", descriptor=("remote-yaml", str(context.settings_path))),
            probe=lambda binding: binding.descriptor,
            load=lambda binding: LoadedSource(
                token="token",
                payload={"app": {"name": "remote"}},
            ),
            validate_final_binding=None,
        )
    )

    missing_file = tmp_path / "missing.yaml"
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(missing_file)
    init_settings()
    assert GetSettings(target=AppSettings, field="name") == "remote"


def test_dotenv_path_switch_sync_can_be_enabled_explicitly(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("FASTAPIEX__SETTINGS__RELOAD", "on_change")

    @Settings("app")
    class AppSettings(BaseSettings):
        name: str
        token: str | None = None

    first_dir = tmp_path / "first"
    second_dir = tmp_path / "second"
    first_dir.mkdir()
    second_dir.mkdir()

    first_yaml = first_dir / "settings.yaml"
    second_yaml = second_dir / "settings.yaml"
    (first_dir / ".env").write_text("TEST__APP__TOKEN=token-first\n", encoding="utf-8")
    (second_dir / ".env").write_text("TEST__APP__TOKEN=token-second\n", encoding="utf-8")

    first_yaml.write_text(
        f'app:\n  name: first\nfastapiex:\n  settings:\n    path: "{first_yaml}"\n',
        encoding="utf-8",
    )
    second_yaml.write_text("app:\n  name: second\n", encoding="utf-8")

    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(first_yaml)
    init_settings()
    assert GetSettings(target=AppSettings, field="name") == "first"
    assert GetSettings(target=AppSettings, field="token") == "token-first"

    manager = get_settings_manager()
    builtin_dotenv = manager.get_source("dotenv")
    assert builtin_dotenv is not None
    manager.register_source(
        replace(
            builtin_dotenv,
            policy=replace(
                builtin_dotenv.policy,
                follow_context=True,
            ),
        )
    )

    first_yaml.write_text(
        f'app:\n  name: first-updated\nfastapiex:\n  settings:\n    path: "{second_yaml}"\n',
        encoding="utf-8",
    )

    assert GetSettings(target=AppSettings, field="name") == "second"
    assert GetSettings(target=AppSettings, field="token") == "token-second"


def test_getsettings_allows_mapping_category_injection_for_single_settingsmap(tmp_path: Path) -> None:
    @SettingsMap("services")
    class ServiceSettings(BaseSettings):
        host: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("services:\n  api:\n    host: localhost\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    services = GetSettings(target=dict)
    assert isinstance(services, dict)
    assert services["api"].host == "localhost"


def test_getsettings_mapping_category_injection_requires_unique_map_section(tmp_path: Path) -> None:
    @SettingsMap("services")
    class ServiceSettings(BaseSettings):
        host: str

    @SettingsMap("databases")
    class DbSettings(BaseSettings):
        url: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text(
        "services:\n  api:\n    host: localhost\ndatabases:\n  main:\n    url: sqlite:///main.db\n",
        encoding="utf-8",
    )
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    with pytest.raises(SettingsResolveError, match="matched multiple sections"):
        GetSettings(target=dict)


def test_getsettings_supports_generic_type_injection_when_unique(tmp_path: Path) -> None:
    class ConfigMarker:
        pass

    @Settings("app")
    class AppSettings(BaseSettings, ConfigMarker):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("app:\n  name: demo\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    settings_obj = GetSettings(target=ConfigMarker)
    assert isinstance(settings_obj, AppSettings)
    assert settings_obj.name == "demo"


def test_getsettings_generic_type_injection_requires_unique_section(tmp_path: Path) -> None:
    class SharedMarker:
        pass

    @Settings("app")
    class AppSettings(BaseSettings, SharedMarker):
        name: str

    @Settings("worker")
    class WorkerSettings(BaseSettings, SharedMarker):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text(
        "app:\n  name: api\nworker:\n  name: jobs\n",
        encoding="utf-8",
    )
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    with pytest.raises(SettingsResolveError, match="matched multiple sections"):
        GetSettings(target=SharedMarker)


def test_getsettings_generic_type_ambiguity_falls_back_to_default(tmp_path: Path) -> None:
    class SharedMarker:
        pass

    @Settings("app")
    class AppSettings(BaseSettings, SharedMarker):
        name: str

    @Settings("worker")
    class WorkerSettings(BaseSettings, SharedMarker):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text(
        "app:\n  name: api\nworker:\n  name: jobs\n",
        encoding="utf-8",
    )
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    default_obj = types.SimpleNamespace(name="fallback")
    assert GetSettings(target=SharedMarker, default=default_obj) is default_obj


def test_getsettings_generic_type_miss_falls_back_to_default(tmp_path: Path) -> None:
    class KnownMarker:
        pass

    class MissingMarker:
        pass

    @Settings("app")
    class AppSettings(BaseSettings, KnownMarker):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("app:\n  name: api\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    default_obj = types.SimpleNamespace(name="fallback")
    assert GetSettings(target=MissingMarker, default=default_obj) is default_obj


def test_getsettings_supports_generic_type_injection_for_singleton_map_item(tmp_path: Path) -> None:
    class ServiceMarker:
        pass

    @SettingsMap("services")
    class ServiceSettings(BaseSettings, ServiceMarker):
        host: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("services:\n  api:\n    host: localhost\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    services = GetSettings(target=ServiceMarker)
    assert isinstance(services, dict)
    assert services["api"].host == "localhost"


def test_getsettings_can_read_intermediate_path_for_object_and_map_levels(tmp_path: Path) -> None:
    @Settings("father")
    class FatherSettings(BaseSettings):
        a: int

    @Settings("father.son.grandson")
    class GrandsonSettings(BaseSettings):
        age: int

    @SettingsMap("father.services")
    class ServiceSettings(BaseSettings):
        host: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text(
        "father:\n  a: 1\n  son:\n    grandson:\n      age: 7\n  services:\n    api:\n      host: 127.0.0.1\n",
        encoding="utf-8",
    )
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    son = GetSettings(target="father.son")
    assert son.grandson.age == 7

    services = GetSettings(target="father.services")
    assert isinstance(services, dict)
    assert services["api"].host == "127.0.0.1"


def test_default_with_field_returns_default_object_instead_of_projecting() -> None:
    default_obj = types.SimpleNamespace(val="test")
    assert GetSettings(target="some_unknown_path", field="val", default=default_obj) is default_obj


def test_manual_reload_restores_snapshot_after_monkeypatch(tmp_path: Path) -> None:
    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("fastapiex:\n  settings:\n    reload: off\napp:\n  name: stable\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    app = GetSettings(target=AppSettings)
    app.name = "patched"
    assert GetSettings(target=AppSettings, field="name") == "patched"

    reload_settings(reason="restore-after-monkeypatch")
    assert GetSettings(target=AppSettings, field="name") == "stable"


def test_unrelated_module_changes_do_not_force_refresh_when_reload_off(
    tmp_path: Path,
) -> None:
    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("app:\n  name: v1\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    assert GetSettings(target=AppSettings, field="name") == "v1"

    settings_file.write_text("app:\n  name: v2\n", encoding="utf-8")
    assert GetSettings(target=AppSettings, field="name") == "v1"

    dynamic_name = "tests.module_change_trigger"
    sys.modules[dynamic_name] = types.ModuleType(dynamic_name)
    try:
        assert GetSettings(target=AppSettings, field="name") == "v1"
    finally:
        sys.modules.pop(dynamic_name, None)


def test_module_lifecycle_removes_stale_section_after_module_replacement(tmp_path: Path) -> None:
    module_v1 = types.ModuleType(DYNAMIC_MODULE)
    sys.modules[DYNAMIC_MODULE] = module_v1
    exec(
        """
from fastapiex.settings import BaseSettings, Settings

@Settings("dyn")
class DynSettings(BaseSettings):
    a: int
""",
        module_v1.__dict__,
    )

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("dyn:\n  a: 5\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    assert GetSettings(target="dyn", field="a") == 5

    # Replace module with a same-name module that no longer declares settings.
    module_v2 = types.ModuleType(DYNAMIC_MODULE)
    module_v2.__dict__["__name__"] = DYNAMIC_MODULE
    sys.modules[DYNAMIC_MODULE] = module_v2

    # phase-A miss should trigger lifecycle reconciliation and remove stale declarations.
    assert GetSettings(target="dyn", field="missing", default=-1) == -1

    with pytest.raises(SettingsResolveError):
        GetSettings(target="dyn", field="a")


def test_module_lifecycle_reconcile_handles_mid_scan_module_replacement(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    module_v1 = types.ModuleType(DYNAMIC_MODULE)
    sys.modules[DYNAMIC_MODULE] = module_v1
    exec(
        """
from fastapiex.settings import BaseSettings, Settings

@Settings("dyn")
class DynSettings(BaseSettings):
    a: int
""",
        module_v1.__dict__,
    )

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("dyn:\n  a: 13\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()
    assert GetSettings(target="dyn", field="a") == 13

    registry = registry_module.get_settings_registry()
    original_check = registry._record_is_live_locked
    replaced = False

    def _replace_module_before_liveness_check(*, model: type[BaseSettings], record: object) -> bool:
        nonlocal replaced
        if not replaced:
            module_v2 = types.ModuleType(DYNAMIC_MODULE)
            module_v2.__dict__["__name__"] = DYNAMIC_MODULE
            sys.modules[DYNAMIC_MODULE] = module_v2
            replaced = True
        return original_check(model=model, record=record)

    monkeypatch.setattr(registry, "_record_is_live_locked", _replace_module_before_liveness_check)

    assert GetSettings(target="dyn", field="missing", default=-1) == -1
    with pytest.raises(SettingsResolveError):
        GetSettings(target="dyn", field="a")


def test_module_lifecycle_removes_stale_section_after_inplace_module_reset(tmp_path: Path) -> None:
    module = types.ModuleType(DYNAMIC_MODULE)
    sys.modules[DYNAMIC_MODULE] = module
    exec(
        """
from fastapiex.settings import BaseSettings, Settings

@Settings("dyn")
class DynSettings(BaseSettings):
    a: int
""",
        module.__dict__,
    )

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("dyn:\n  a: 7\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    assert GetSettings(target="dyn", field="a") == 7

    module.__dict__.clear()
    module.__dict__["__name__"] = DYNAMIC_MODULE

    assert GetSettings(target="dyn", field="missing", default=-1) == -1
    with pytest.raises(SettingsResolveError):
        GetSettings(target="dyn", field="a")


def test_same_module_identity_redefinition_replaces_declared_model(tmp_path: Path) -> None:
    module = types.ModuleType(DYNAMIC_MODULE)
    sys.modules[DYNAMIC_MODULE] = module

    exec(
        """
from fastapiex.settings import BaseSettings, Settings

@Settings("dyn")
class DynSettings(BaseSettings):
    a: int = 1
""",
        module.__dict__,
    )

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("{}\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()
    assert GetSettings(target="dyn", field="a") == 1

    exec(
        """
from fastapiex.settings import BaseSettings, Settings

@Settings("dyn")
class DynSettings(BaseSettings):
    b: int = 2
""",
        module.__dict__,
    )

    assert GetSettings(target="dyn", field="b") == 2
    assert GetSettings(target="dyn", field="a", default=-1) == -1


def test_reload_reconciles_module_unload_and_drops_stale_section(tmp_path: Path) -> None:
    module = types.ModuleType(DYNAMIC_MODULE)
    sys.modules[DYNAMIC_MODULE] = module
    exec(
        """
from fastapiex.settings import BaseSettings, Settings

@Settings("dyn")
class DynSettings(BaseSettings):
    a: int
""",
        module.__dict__,
    )

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("dyn:\n  a: 11\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()
    assert GetSettings(target="dyn", field="a") == 11

    sys.modules.pop(DYNAMIC_MODULE, None)
    reload_settings(reason="module-unload")

    with pytest.raises(SettingsResolveError):
        GetSettings(target="dyn", field="a")


def test_cached_miss_is_invalidated_when_new_declaration_registers(tmp_path: Path) -> None:
    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("{}\n", encoding="utf-8")
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    assert GetSettings(target="dyn", field="a", default=-1) == -1

    module = types.ModuleType(DYNAMIC_MODULE)
    sys.modules[DYNAMIC_MODULE] = module
    try:
        exec(
            """
from fastapiex.settings import BaseSettings, Settings

@Settings("dyn")
class DynSettings(BaseSettings):
    a: int = 7
""",
            module.__dict__,
        )
        assert GetSettings(target="dyn", field="a") == 7
    finally:
        sys.modules.pop(DYNAMIC_MODULE, None)


def test_settings_source_is_process_global_and_singleton(tmp_path: Path) -> None:
    @Settings("app")
    class AppSettings(BaseSettings):
        name: str = "demo"

    file_a = tmp_path / "a.yaml"
    file_a.write_text("app:\n  name: a\n", encoding="utf-8")
    file_b = tmp_path / "b.yaml"
    file_b.write_text("app:\n  name: b\n", encoding="utf-8")

    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(file_a)
    init_settings()

    with pytest.raises(RuntimeError, match="different source"):
        os.environ["FASTAPIEX__SETTINGS__PATH"] = str(file_b)
        init_settings()


def test_reload_true_startup_follows_multi_hop_settings_path_chain(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("FASTAPIEX__SETTINGS__RELOAD", "true")

    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    first = tmp_path / "settings.yaml"
    second = tmp_path / "second.yaml"
    third = tmp_path / "third.yaml"

    first.write_text(
        f'fastapiex:\n  settings:\n    path: "{second}"\napp:\n  name: first\n',
        encoding="utf-8",
    )
    second.write_text(
        f'fastapiex:\n  settings:\n    path: "{third}"\napp:\n  name: second\n',
        encoding="utf-8",
    )
    third.write_text("app:\n  name: third\n", encoding="utf-8")

    monkeypatch.chdir(tmp_path)
    init_settings()

    manager = get_settings_manager()
    runtime_view = manager.inspect_runtime()
    assert runtime_view is not None
    assert runtime_view.context.settings_path == third
    assert GetSettings(target=AppSettings, field="name") == "third"


def test_reload_true_runtime_yaml_change_can_trigger_multi_hop_path_switch(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("FASTAPIEX__SETTINGS__RELOAD", "true")

    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    first = tmp_path / "first.yaml"
    second = tmp_path / "second.yaml"
    third = tmp_path / "third.yaml"

    first.write_text("app:\n  name: first\n", encoding="utf-8")
    second.write_text(
        f'fastapiex:\n  settings:\n    path: "{third}"\napp:\n  name: second\n',
        encoding="utf-8",
    )
    third.write_text("app:\n  name: third\n", encoding="utf-8")

    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(first)
    init_settings()
    assert GetSettings(target=AppSettings, field="name") == "first"

    first.write_text(
        f'fastapiex:\n  settings:\n    path: "{second}"\napp:\n  name: first-updated\n',
        encoding="utf-8",
    )

    manager = get_settings_manager()
    assert GetSettings(target=AppSettings, field="name") == "third"
    runtime_view = manager.inspect_runtime()
    assert runtime_view is not None
    assert runtime_view.context.settings_path == third


def test_fastapiex_controls_with_mixed_case_across_yaml_dotenv_env_follow_source_precedence(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("fAsTaPiEx__SeTtInGs__ReLoAd", "true")
    monkeypatch.setenv("FASTAPIEX__SETTINGS__CASE_SENSITIVE", "false")

    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text(
        "FastAPIEx:\n  Settings:\n    Reload: off\n    Case_Sensitive: true\napp:\n  name: yaml-v1\n",
        encoding="utf-8",
    )
    (tmp_path / ".env").write_text(
        "FASTAPIEX__SETTINGS__RELOAD=always\nFASTAPIEX__SETTINGS__CASE_SENSITIVE=true\n",
        encoding="utf-8",
    )

    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()
    assert GetSettings(target="FASTAPIEX.SETTINGS.RELOAD") is True
    assert GetSettings(target="fastapiex.settings.case_sensitive") is False
    assert GetSettings(target=AppSettings, field="name") == "yaml-v1"

    settings_file.write_text(
        "FastAPIEx:\n  Settings:\n    Reload: off\n    Case_Sensitive: true\napp:\n  name: yaml-v2\n",
        encoding="utf-8",
    )
    assert GetSettings(target=AppSettings, field="name") == "yaml-v2"


def test_business_nested_fastapiex_path_remains_business_data_in_case_sensitive_mode(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("FASTAPIEX__SETTINGS__CASE_SENSITIVE", "true")

    @Settings("app")
    class AppSettings(BaseSettings):
        fastapiex: dict[str, dict[str, str]]

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text(
        "app:\n  fastapiex:\n    Settings:\n      TokenKey: Value\n",
        encoding="utf-8",
    )
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    assert GetSettings(target=AppSettings, field="fastapiex.Settings.TokenKey") == "Value"
    assert GetSettings(target="app.fastapiex.settings.tokenkey", default="miss") == "miss"
    assert GetSettings(target="FASTAPIEX.SETTINGS.CASE_SENSITIVE") is True


def test_case_insensitive_mode_handles_extreme_mixed_case_sections_fields_and_maps(
    tmp_path: Path,
) -> None:
    @Settings("CoRe.App")
    class AppSettings(BaseSettings):
        title: str

    @SettingsMap("MiXeD.Services")
    class ServiceSettings(BaseSettings):
        host: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text(
        "cOrE:\n  aPp:\n    TiTlE: Demo\nmIxEd:\n  sErViCeS:\n    ApI:\n      HoSt: 127.0.0.1\n",
        encoding="utf-8",
    )
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    assert GetSettings(target="CORE.APP.TITLE") == "Demo"
    assert GetSettings(target=AppSettings, field="TITLE") == "Demo"
    assert GetSettings(target="mixed.services.api.host") == "127.0.0.1"
    assert GetSettings(target=ServiceSettings, field="Api.Host") == "127.0.0.1"


def test_case_sensitive_mode_distinguishes_business_key_variants_but_controls_stay_insensitive(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("FASTAPIEX__SETTINGS__CASE_SENSITIVE", "true")
    monkeypatch.setenv("fAsTaPiEx__SeTtInGs__ReLoAd", "off")

    @Settings("App")
    class AppTitle(BaseSettings):
        name: str

    @Settings("APP")
    class UpperAppTitle(BaseSettings):
        name: str

    @Settings("app")
    class LowerAppTitle(BaseSettings):
        name: str

    @SettingsMap("Services")
    class ServiceSettings(BaseSettings):
        host: str

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text(
        "App:\n  name: mixed\nAPP:\n  name: upper\napp:\n  name: lower\n"
        "Services:\n  API:\n    host: upper-host\n  api:\n    host: lower-host\n",
        encoding="utf-8",
    )
    os.environ["FASTAPIEX__SETTINGS__PATH"] = str(settings_file)
    init_settings()

    assert GetSettings(target="App", field="name") == "mixed"
    assert GetSettings(target="APP", field="name") == "upper"
    assert GetSettings(target="app", field="name") == "lower"
    assert GetSettings(target=AppTitle, field="name") == "mixed"
    assert GetSettings(target=UpperAppTitle, field="name") == "upper"
    assert GetSettings(target=LowerAppTitle, field="name") == "lower"

    assert GetSettings(target="Services.API.host") == "upper-host"
    assert GetSettings(target="Services.api.host") == "lower-host"
    services = GetSettings(target=ServiceSettings)
    assert isinstance(services, dict)
    assert set(services.keys()) == {"API", "api"}
    assert GetSettings(target="FASTAPIEX.SETTINGS.RELOAD") is False

    with pytest.raises(SettingsResolveError):
        GetSettings(target="aPP", field="name")


def test_combined_extreme_controls_path_hops_and_business_fastapiex_collision(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("FASTAPIEX__SETTINGS__RELOAD", "true")
    monkeypatch.setenv("fAsTaPiEx__SeTtInGs__CaSe_SeNsItIvE", "false")

    @Settings("App")
    class AppSettings(BaseSettings):
        name: str
        fastapiex: dict[str, dict[str, str]]

    @SettingsMap("Services")
    class ServiceSettings(BaseSettings):
        host: str

    hop1 = tmp_path / "hop1"
    hop2 = tmp_path / "hop2"
    hop3 = tmp_path / "hop3"
    hop4 = tmp_path / "hop4"
    hop1.mkdir()
    hop2.mkdir()
    hop3.mkdir()
    hop4.mkdir()

    first = hop1 / "settings.yaml"
    second = hop2 / "settings.yaml"
    third = hop3 / "settings.yaml"
    fourth = hop4 / "settings.yaml"

    (hop1 / ".env").write_text(
        "FASTAPIEX__SETTINGS__CASE_SENSITIVE=true\nFASTAPIEX__SETTINGS__RELOAD=off\n",
        encoding="utf-8",
    )
    first.write_text(
        f"FastAPIEx:\n  Settings:\n    Path: \"{second}\"\n"
        "App:\n  Name: hop1\n  FASTAPIEX:\n    Inner:\n      ToKeN: alpha\n"
        "Services:\n  Api:\n    Host: hop1-host\n",
        encoding="utf-8",
    )
    second.write_text(
        f"fAsTaPiEx:\n  sEtTiNgS:\n    pAtH: \"{hop3}\"\n"
        "app:\n  name: hop2\n  fastapiex:\n    inner:\n      token: beta\n"
        "services:\n  api:\n    host: hop2-host\n",
        encoding="utf-8",
    )
    third.write_text(
        "FASTAPIEX:\n  SETTINGS:\n    Reload: always\n"
        "App:\n  Name: hop3\n  FastAPIEx:\n    InNeR:\n      ToKeN: gamma\n"
        "Services:\n  Api:\n    Host: hop3-host\n",
        encoding="utf-8",
    )
    fourth.write_text(
        "app:\n  name: hop4\n  fastapiex:\n    inner:\n      token: delta\n"
        "services:\n  api:\n    host: hop4-host\n",
        encoding="utf-8",
    )

    monkeypatch.chdir(hop1)
    init_settings()
    assert GetSettings(target=AppSettings, field="name") == "hop3"
    assert GetSettings(target="app.fastapiex.inner.token") == "gamma"
    assert GetSettings(target="services.api.host") == "hop3-host"
    assert GetSettings(target="FASTAPIEX.SETTINGS.CASE_SENSITIVE") is False
    # Path-switch sync re-reads yaml with newer revisions; LWW lets newer yaml override older env values.
    assert GetSettings(target="FASTAPIEX.SETTINGS.RELOAD") == "always"

    third.write_text(
        f"FASTAPIEX:\n  SETTINGS:\n    PATH: \"{fourth}\"\n"
        "App:\n  Name: hop3-updated\n  FastAPIEx:\n    InNeR:\n      ToKeN: gamma2\n"
        "Services:\n  Api:\n    Host: hop3-updated-host\n",
        encoding="utf-8",
    )
    assert GetSettings(target=AppSettings, field="name") == "hop4"
    assert GetSettings(target="APP.FASTAPIEX.INNER.TOKEN") == "delta"
    assert GetSettings(target=ServiceSettings, field="api.host") == "hop4-host"


def test_reload_true_with_settings_path_cycle_warns_and_stops_at_current_source(
    caplog: pytest.LogCaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("FASTAPIEX__SETTINGS__RELOAD", "true")
    caplog.set_level(logging.WARNING)

    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    first = tmp_path / "settings.yaml"
    second = tmp_path / "second.yaml"
    first.write_text(
        f'fastapiex:\n  settings:\n    path: "{second}"\napp:\n  name: first\n',
        encoding="utf-8",
    )
    second.write_text(
        f'fastapiex:\n  settings:\n    path: "{first}"\napp:\n  name: second\n',
        encoding="utf-8",
    )

    monkeypatch.chdir(tmp_path)
    init_settings()

    manager = get_settings_manager()
    runtime_view = manager.inspect_runtime()
    assert runtime_view is not None
    assert runtime_view.context.settings_path == first
    assert GetSettings(target=AppSettings, field="name") == "first"

    warning_messages = [record.getMessage() for record in caplog.records if record.levelno >= logging.WARNING]
    assert any("path control cycle detected" in message for message in warning_messages)


def test_path_cycle_detection_keys_by_settings_path_not_entire_source(
    caplog: pytest.LogCaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    monkeypatch.setenv("FASTAPIEX__SETTINGS__RELOAD", "true")
    monkeypatch.delenv("FASTAPIEX__SETTINGS__ENV_PREFIX", raising=False)
    caplog.set_level(logging.WARNING)

    @Settings("app")
    class AppSettings(BaseSettings):
        name: str

    first = tmp_path / "settings.yaml"
    second = tmp_path / "second.yaml"
    first.write_text(
        f'fastapiex:\n  settings:\n    path: "{second}"\napp:\n  name: first\n',
        encoding="utf-8",
    )
    second.write_text(
        f'fastapiex:\n  settings:\n    path: "{first}"\napp:\n  name: second\n',
        encoding="utf-8",
    )

    monkeypatch.chdir(tmp_path)
    init_settings()

    manager = get_settings_manager()
    runtime_view = manager.inspect_runtime()
    assert runtime_view is not None
    assert runtime_view.context.settings_path == first
    assert GetSettings(target=AppSettings, field="name") == "first"

    warning_messages = [record.getMessage() for record in caplog.records if record.levelno >= logging.WARNING]
    assert any("path control cycle detected" in message for message in warning_messages)



from __future__ import annotations

import logging
from collections.abc import Callable, Mapping
from dataclasses import dataclass
from typing import Any

from .context import ConfigContext
from .live_config import EntrySource, build_entries_from_mappings
from .projection import materialize_control_snapshot
from .runtime_state import SourceSnapshot
from .source_contract import SourceRegistry, SourceSpec
from .types import SourceSyncMode

_MAX_CONTROL_HOPS = 16


@dataclass(frozen=True)
class CandidateRuntime:
    last_rev: int
    context: ConfigContext
    snapshots: dict[str, SourceSnapshot]
    changed: bool


def build_candidate_runtime(
    *,
    initial_context: ConfigContext,
    current_snapshots: Mapping[str, SourceSnapshot],
    current_last_rev: int,
    mode: SourceSyncMode,
    sources: SourceRegistry,
    build_context_from_controls: Callable[[Mapping[str, Any], ConfigContext], ConfigContext],
    logger: logging.Logger,
) -> CandidateRuntime:
    context = initial_context
    snapshots = dict(current_snapshots)
    last_rev = current_last_rev
    changed = False
    visited_targets: set[tuple[object, object]] = {context.cycle_key()}

    for _ in range(_MAX_CONTROL_HOPS):
        refreshed_snapshots, refreshed_last_rev, refreshed = refresh_snapshot_set(
            mode=mode,
            context=context,
            current=snapshots,
            current_last_rev=last_rev,
            sources=sources,
        )
        if not refreshed and not changed and context == initial_context:
            return CandidateRuntime(
                last_rev=refreshed_last_rev,
                context=context,
                snapshots=refreshed_snapshots,
                changed=False,
            )

        control_snapshot = materialize_control_snapshot(
            build_entries_from_runtime_snapshots(refreshed_snapshots, sources=sources)
        )
        next_context = build_context_from_controls(control_snapshot, context)

        if next_context.cycle_key() != context.cycle_key():
            if next_context.cycle_key() in visited_targets:
                logger.warning(
                    "settings path control cycle detected; keeping path=%s",
                    context.settings_path,
                )
                return CandidateRuntime(
                    last_rev=refreshed_last_rev,
                    context=context,
                    snapshots=refreshed_snapshots,
                    changed=True,
                )

            visited_targets.add(next_context.cycle_key())
            context = next_context
            snapshots = refreshed_snapshots
            last_rev = refreshed_last_rev
            changed = True
            continue

        if next_context != context:
            context = next_context
            snapshots = refreshed_snapshots
            last_rev = refreshed_last_rev
            changed = True
            continue

        return CandidateRuntime(
            last_rev=refreshed_last_rev,
            context=context,
            snapshots=refreshed_snapshots,
            changed=changed or refreshed or context != initial_context,
        )

    logger.warning("settings control convergence hop limit reached; keeping path=%s", context.settings_path)
    return CandidateRuntime(
        last_rev=last_rev,
        context=context,
        snapshots=snapshots,
        changed=True,
    )


def refresh_snapshot_set(
    *,
    mode: SourceSyncMode,
    context: ConfigContext,
    current: Mapping[str, SourceSnapshot],
    current_last_rev: int,
    sources: SourceRegistry,
) -> tuple[dict[str, SourceSnapshot], int, bool]:
    if mode == "none":
        return dict(current), current_last_rev, False

    if mode == "auto" and context.reload_mode == "off":
        return dict(current), current_last_rev, False

    next_snapshots: dict[str, SourceSnapshot] = {}
    changed_snapshots: list[tuple[SourceSpec, SourceSnapshot]] = []

    for spec in sources.ordered():
        previous = current.get(spec.name)
        binding = (
            previous.binding
            if previous is not None and not spec.policy.follow_context and mode != "full"
            else spec.bind(context)
        )
        if not _should_load_source(
            spec=spec,
            context=context,
            previous=previous,
            binding=binding,
            mode=mode,
        ):
            if previous is not None:
                next_snapshots[spec.name] = previous
            continue

        loaded = spec.load(binding)
        previous_rev = 0 if previous is None else previous.rev
        snapshot = SourceSnapshot(
            source=spec.name,
            rev=previous_rev,
            binding=binding,
            token=loaded.token,
            payload=loaded.payload,
        )
        if previous == snapshot:
            if previous is not None:
                next_snapshots[spec.name] = previous
            continue
        changed_snapshots.append((spec, snapshot))

    next_last_rev = current_last_rev
    for offset, (spec, snapshot) in enumerate(changed_snapshots, start=1):
        next_last_rev = current_last_rev + offset
        next_snapshots[spec.name] = SourceSnapshot(
            source=spec.name,
            rev=next_last_rev,
            binding=snapshot.binding,
            token=snapshot.token,
            payload=snapshot.payload,
        )

    return next_snapshots, next_last_rev, bool(changed_snapshots)


def build_entries_from_runtime_snapshots(
    snapshots: Mapping[str, SourceSnapshot],
    *,
    sources: SourceRegistry,
) -> tuple[Any, ...]:
    rows: list[EntrySource] = []
    for spec in sources.ordered():
        snapshot = snapshots.get(spec.name)
        if snapshot is None:
            continue
        rows.append(
            EntrySource(
                source=spec.name,
                priority=spec.priority,
                kind=spec.projection_kind,
                include_in_control=spec.policy.participates_in_controls,
                rev=snapshot.rev,
                mapping=snapshot.payload,
            )
        )
    return build_entries_from_mappings(rows)


def validate_final_source_bindings(
    *,
    context: ConfigContext,
    snapshots: Mapping[str, SourceSnapshot],
    sources: SourceRegistry,
) -> None:
    for spec in sources.ordered():
        if spec.validate_final_binding is None:
            continue
        snapshot = snapshots.get(spec.name)
        binding = spec.bind(context) if snapshot is None else snapshot.binding
        spec.validate_final_binding(context, binding)


def _should_load_source(
    *,
    spec: SourceSpec,
    context: ConfigContext,
    previous: SourceSnapshot | None,
    binding: Any,
    mode: SourceSyncMode,
) -> bool:
    if previous is None:
        return True

    if binding != previous.binding and spec.policy.follow_context:
        return True

    if mode == "full":
        return True

    if mode == "reload":
        return spec.policy.manual_refresh

    if mode != "auto":
        return False

    if context.reload_mode == "always":
        return spec.policy.auto_refresh

    if not spec.policy.auto_refresh:
        return False

    probe_token = spec.probe(binding)
    return probe_token != previous.token


__all__ = [
    "build_candidate_runtime",
    "build_entries_from_runtime_snapshots",
    "CandidateRuntime",
    "refresh_snapshot_set",
    "validate_final_source_bindings",
]



from __future__ import annotations

from collections.abc import Callable, Iterable, Mapping
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Protocol, get_args, get_origin

from pydantic import BaseModel

from .constants import ENV_KEY_SEPARATOR
from .control_contract import CONTROL_ENV_PREFIX, CONTROL_ROOT, CONTROL_SPEC
from .live_config import SourceEntry
from .loader import key_to_parts, parse_env_value

WinnerMeta = tuple[int, int, Any]
_ProjectedEntry = tuple[tuple[str, ...], Any]
_Projector = Callable[[SourceEntry], _ProjectedEntry | None]
_PathResolver = Callable[[str], tuple[str, ...] | None]


def set_nested_force(target: dict[str, Any], path: tuple[str, ...], value: Any) -> None:
    cursor = target
    for part in path[:-1]:
        existing = cursor.get(part)
        if not isinstance(existing, dict):
            existing = {}
            cursor[part] = existing
        cursor = existing
    cursor[path[-1]] = value


def build_snapshot_from_winners(winners: Mapping[tuple[str, ...], WinnerMeta]) -> dict[str, Any]:
    merged: dict[str, Any] = {}
    ordered = sorted(
        winners.items(),
        key=lambda item: (item[1][0], item[1][1], len(item[0]), item[0]),
    )
    for path, (_, _, value) in ordered:
        set_nested_force(merged, path, deepcopy(value))
    return merged


def merge_nested_mapping(target: dict[str, Any], incoming: Mapping[str, Any]) -> None:
    for key, value in incoming.items():
        existing = target.get(key)
        if isinstance(existing, dict) and isinstance(value, Mapping):
            merge_nested_mapping(existing, value)
            continue
        target[key] = deepcopy(value)


def assign_projected_value(target: dict[str, Any], key: str, value: Any) -> None:
    existing = target.get(key)
    if isinstance(existing, dict) and isinstance(value, Mapping):
        merge_nested_mapping(existing, value)
        return
    target[key] = deepcopy(value)


def normalize_control_mapping(raw: Mapping[Any, Any]) -> dict[str, Any]:
    projected: dict[str, Any] = {}
    _merge_casefold_mapping(projected, raw, deepcopy_values=True)
    return projected


def _merge_casefold_mapping(
    target: dict[str, Any],
    incoming: Mapping[Any, Any],
    *,
    deepcopy_values: bool,
) -> None:
    for key, value in incoming.items():
        if not isinstance(key, str):
            continue
        canonical_key = key.casefold()
        if isinstance(value, Mapping):
            existing = target.get(canonical_key)
            nested: dict[str, Any]
            if isinstance(existing, dict):
                nested = existing
            else:
                nested = {}
                target[canonical_key] = nested
            _merge_casefold_mapping(nested, value, deepcopy_values=deepcopy_values)
            continue
        target[canonical_key] = deepcopy(value) if deepcopy_values else value


class _ProjectionPolicy(Protocol):
    def project(self, entry: SourceEntry) -> _ProjectedEntry | None: ...


@dataclass(frozen=True)
class _ControlProjectionPolicy:
    control_path: tuple[str, ...] = CONTROL_SPEC.path
    control_env_prefix: str = CONTROL_ENV_PREFIX

    def project(self, entry: SourceEntry) -> _ProjectedEntry | None:
        if not entry.include_in_control:
            return None
        if entry.kind == "mapping":
            return self._project_yaml(entry)
        return _project_env_entry(entry, key_to_path=self._control_env_key_to_path)

    def _project_yaml(self, entry: SourceEntry) -> _ProjectedEntry | None:
        if not entry.path:
            return None
        if len(entry.path) < len(self.control_path):
            return None
        entry_prefix = entry.path[: len(self.control_path)]
        if any(
            left.casefold() != right.casefold()
            for left, right in zip(entry_prefix, self.control_path, strict=False)
        ):
            return None
        canonical_path = tuple(segment.casefold() for segment in entry.path)
        return (canonical_path, entry.value)

    def _control_env_key_to_path(self, env_key: str) -> tuple[str, ...] | None:
        if not env_key.upper().startswith(self.control_env_prefix):
            return None
        raw_parts = env_key.split(ENV_KEY_SEPARATOR)
        if any(not part for part in raw_parts):
            return None
        return tuple(part.casefold() for part in raw_parts)


@dataclass(frozen=True)
class _SettingsProjectionPolicy:
    env_prefix: str
    case_sensitive: bool

    def project(self, entry: SourceEntry) -> _ProjectedEntry | None:
        if entry.kind == "mapping":
            return _project_yaml_entry(entry)
        return _project_env_entry(entry, key_to_path=self._settings_env_key_to_path)

    def _settings_env_key_to_path(self, env_key: str) -> tuple[str, ...] | None:
        parts = key_to_parts(env_key, prefix=self.env_prefix, case_sensitive=self.case_sensitive)
        if parts is None:
            return None
        return tuple(parts)


_CONTROL_POLICY = _ControlProjectionPolicy()


def materialize_control_snapshot(entries: Iterable[SourceEntry]) -> dict[str, Any]:
    return _materialize_snapshot(entries, policy=_CONTROL_POLICY)


def materialize_effective_snapshot(
    entries: Iterable[SourceEntry],
    *,
    env_prefix: str,
    case_sensitive: bool,
) -> dict[str, Any]:
    policy = _SettingsProjectionPolicy(env_prefix=env_prefix, case_sensitive=case_sensitive)
    return _materialize_snapshot(entries, policy=policy)


def _materialize_snapshot(entries: Iterable[SourceEntry], *, policy: _ProjectionPolicy) -> dict[str, Any]:
    winners = _collect_projected_winners(entries, projector=policy.project)
    return build_snapshot_from_winners(winners)


def _collect_projected_winners(
    entries: Iterable[SourceEntry],
    *,
    projector: _Projector,
) -> dict[tuple[str, ...], WinnerMeta]:
    winners: dict[tuple[str, ...], WinnerMeta] = {}
    for entry in entries:
        projected = projector(entry)
        if projected is None:
            continue

        path, value = projected
        meta = (entry.rev, entry.priority)
        existing = winners.get(path)
        if existing is not None and meta <= (existing[0], existing[1]):
            continue
        winners[path] = (meta[0], meta[1], deepcopy(value))
    return winners


def _project_yaml_entry(entry: SourceEntry) -> _ProjectedEntry | None:
    if not entry.path:
        return None
    return (entry.path, entry.value)


def _project_env_entry(
    entry: SourceEntry,
    *,
    key_to_path: _PathResolver,
) -> _ProjectedEntry | None:
    env_key = _entry_env_key(entry)
    if env_key is None:
        return None

    path = key_to_path(env_key)
    if path is None:
        return None

    return (path, _parse_env_like_value(entry.value))


def _entry_env_key(entry: SourceEntry) -> str | None:
    if len(entry.path) != 1:
        return None
    return entry.path[0]


def _parse_env_like_value(value: Any) -> Any:
    if isinstance(value, str):
        return parse_env_value(value)
    return deepcopy(value)


def project_snapshot_for_validation(
    raw: Mapping[str, Any],
    *,
    root_model: type[BaseModel],
    case_sensitive: bool,
) -> dict[str, Any]:
    return _project_mapping_to_model(
        raw,
        root_model=root_model,
        case_sensitive=case_sensitive,
        allow_control_root=True,
    )


def _project_mapping_to_model(
    raw: Mapping[Any, Any],
    *,
    root_model: type[BaseModel],
    case_sensitive: bool,
    allow_control_root: bool = False,
) -> dict[str, Any]:
    projected: dict[str, Any] = {}
    fields = root_model.model_fields

    for key, value in raw.items():
        target_key, projected_value = _project_entry_to_model(
            fields,
            key=key,
            value=value,
            case_sensitive=case_sensitive,
            allow_control_root=allow_control_root,
        )
        assign_projected_value(projected, target_key, projected_value)

    return projected


def _project_entry_to_model(
    fields: Mapping[str, Any],
    *,
    key: Any,
    value: Any,
    case_sensitive: bool,
    allow_control_root: bool,
) -> tuple[str, Any]:
    input_key = key if isinstance(key, str) else str(key)
    control_entry = _project_control_root_entry(
        fields,
        input_key,
        value,
        allow_control_root=allow_control_root,
    )
    if control_entry is not None:
        return control_entry

    field_name = _resolve_field_name(fields, input_key, case_sensitive=case_sensitive)
    if field_name is None:
        return input_key, deepcopy(value)

    return field_name, _project_model_field_value(
        annotation=fields[field_name].annotation,
        value=value,
        case_sensitive=case_sensitive,
    )


def _project_model_field_value(
    *,
    annotation: Any,
    value: Any,
    case_sensitive: bool,
) -> Any:
    object_model = _extract_model_type(annotation)
    if object_model is not None and isinstance(value, Mapping):
        return _project_mapping_to_model(
            value,
            root_model=object_model,
            case_sensitive=case_sensitive,
            allow_control_root=False,
        )

    map_value_model = _extract_mapping_value_model(annotation)
    if map_value_model is not None and isinstance(value, Mapping):
        return _project_mapping_values_to_model(
            value,
            value_model=map_value_model,
            case_sensitive=case_sensitive,
        )

    return deepcopy(value)


def _project_mapping_values_to_model(
    raw: Mapping[Any, Any],
    *,
    value_model: type[BaseModel],
    case_sensitive: bool,
) -> dict[str, Any]:
    projected: dict[str, Any] = {}
    for key, value in raw.items():
        item_key = key if isinstance(key, str) else str(key)
        if isinstance(value, Mapping):
            projected[item_key] = _project_mapping_to_model(
                value,
                root_model=value_model,
                case_sensitive=case_sensitive,
                allow_control_root=False,
            )
            continue
        projected[item_key] = deepcopy(value)
    return projected


def _project_control_root_entry(
    fields: Mapping[str, Any],
    input_key: str,
    value: Any,
    *,
    allow_control_root: bool,
) -> tuple[str, Any] | None:
    if not allow_control_root:
        return None

    control_field = _resolve_control_root_field_name(fields, input_key)
    if control_field is None:
        return None
    if not isinstance(value, Mapping):
        return (control_field, deepcopy(value))
    return (control_field, normalize_control_mapping(value))


def _resolve_control_root_field_name(
    fields: Mapping[str, Any],
    input_key: str,
) -> str | None:
    if input_key.casefold() != CONTROL_ROOT.casefold():
        return None
    matches = [field_name for field_name in fields if field_name.casefold() == CONTROL_ROOT.casefold()]
    if len(matches) != 1:
        return None
    return matches[0]


def _resolve_field_name(
    fields: Mapping[str, Any],
    input_key: str,
    *,
    case_sensitive: bool,
) -> str | None:
    if input_key in fields:
        return input_key

    if case_sensitive:
        return None

    folded = input_key.casefold()
    matches = [field_name for field_name in fields if field_name.casefold() == folded]
    if len(matches) != 1:
        return None

    return matches[0]


def _extract_model_type(annotation: Any) -> type[BaseModel] | None:
    candidate = _unwrap_optional(annotation)
    if isinstance(candidate, type) and issubclass(candidate, BaseModel):
        return candidate
    return None


def _extract_mapping_value_model(annotation: Any) -> type[BaseModel] | None:
    candidate = _unwrap_optional(annotation)
    origin = get_origin(candidate)
    if origin not in {dict, Mapping}:
        return None
    args = get_args(candidate)
    if len(args) != 2:
        return None
    value_type = _unwrap_optional(args[1])
    if isinstance(value_type, type) and issubclass(value_type, BaseModel):
        return value_type
    return None


def _unwrap_optional(annotation: Any) -> Any:
    args = get_args(annotation)
    if not args:
        return annotation
    non_none = [arg for arg in args if arg is not type(None)]
    if len(non_none) == 1 and len(non_none) != len(args):
        return non_none[0]
    return annotation



from __future__ import annotations

from pathlib import Path

from .constants import DOTENV_FILENAME
from .context import ConfigContext
from .exceptions import SettingsValidationError
from .loader import (
    file_state,
    load_dotenv_file_snapshot,
    load_env_snapshot_raw,
    load_yaml_file_snapshot,
)
from .source_contract import LoadedSource, SourceBinding, SourcePolicy, SourceSpec


def builtin_source_specs() -> tuple[SourceSpec, ...]:
    return (
        SourceSpec(
            name="yaml",
            priority=1,
            projection_kind="mapping",
            policy=SourcePolicy(
                auto_refresh=True,
                manual_refresh=True,
                follow_context=True,
            ),
            bind=_bind_yaml,
            probe=_probe_file_binding,
            load=_load_yaml_source,
            validate_final_binding=_validate_yaml_explicit_file,
        ),
        SourceSpec(
            name="dotenv",
            priority=2,
            projection_kind="env_like",
            policy=SourcePolicy(
                auto_refresh=False,
                manual_refresh=False,
                follow_context=False,
            ),
            bind=_bind_dotenv,
            probe=_probe_file_binding,
            load=_load_dotenv_source,
        ),
        SourceSpec(
            name="env",
            priority=3,
            projection_kind="env_like",
            policy=SourcePolicy(
                auto_refresh=False,
                manual_refresh=False,
                follow_context=False,
            ),
            bind=_bind_env,
            probe=_probe_process_env,
            load=_load_env_source,
        ),
    )


def _bind_yaml(context: ConfigContext) -> SourceBinding:
    return SourceBinding(source="yaml", descriptor=context.settings_path)


def _bind_dotenv(context: ConfigContext) -> SourceBinding:
    return SourceBinding(source="dotenv", descriptor=(context.anchor_dir / DOTENV_FILENAME).resolve())


def _bind_env(_: ConfigContext) -> SourceBinding:
    return SourceBinding(source="env", descriptor="process-env")


def _probe_file_binding(binding: SourceBinding) -> object:
    descriptor = binding.descriptor
    if not isinstance(descriptor, Path):
        raise TypeError(f"{binding.source} descriptor must be a Path")
    return file_state(descriptor)


def _load_yaml_source(binding: SourceBinding) -> LoadedSource:
    descriptor = binding.descriptor
    if not isinstance(descriptor, Path):
        raise TypeError("yaml descriptor must be a Path")
    payload, token = load_yaml_file_snapshot(descriptor)
    return LoadedSource(token=token, payload=payload)


def _load_dotenv_source(binding: SourceBinding) -> LoadedSource:
    descriptor = binding.descriptor
    if not isinstance(descriptor, Path):
        raise TypeError("dotenv descriptor must be a Path")
    payload, token = load_dotenv_file_snapshot(descriptor)
    return LoadedSource(token=token, payload=payload)


def _probe_process_env(_: SourceBinding) -> object:
    return tuple(sorted(load_env_snapshot_raw().items()))


def _load_env_source(_: SourceBinding) -> LoadedSource:
    raw = load_env_snapshot_raw()
    return LoadedSource(token=tuple(sorted(raw.items())), payload=raw)


def _validate_yaml_explicit_file(context: ConfigContext, binding: SourceBinding) -> None:
    if context.path_mode != "explicit_file":
        return

    descriptor = binding.descriptor
    if not isinstance(descriptor, Path):
        raise SettingsValidationError(f"explicit settings file is missing: {context.settings_path}")

    path_state = file_state(descriptor)
    exists = False if path_state is None else bool(path_state[1])
    if not exists:
        raise SettingsValidationError(f"explicit settings file is missing: {context.settings_path}")


__all__ = [
    "builtin_source_specs",
]



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from .constants import SETTINGS_FILENAME
from .control_contract import ControlModel
from .loader import resolve_env_prefix
from .types import ReloadMode, SettingsPathMode


@dataclass(frozen=True)
class ResolvedSettingsTarget:
    settings_path: Path
    anchor_dir: Path
    path_mode: SettingsPathMode


@dataclass(frozen=True)
class ConfigContext:
    settings_path: Path
    anchor_dir: Path
    path_mode: SettingsPathMode
    env_prefix: str
    case_sensitive: bool
    reload_mode: ReloadMode

    def cycle_key(self) -> tuple[SettingsPathMode, Path]:
        return (self.path_mode, self.settings_path)


def resolve_settings_target(
    raw: str | Path | None,
    *,
    as_directory: bool = False,
) -> ResolvedSettingsTarget | None:
    if raw is None:
        return None

    if isinstance(raw, Path):
        path = raw.expanduser()
    else:
        text = raw.strip()
        if not text:
            return None
        path = Path(text).expanduser()

    if as_directory:
        anchor_dir = path.resolve()
        return ResolvedSettingsTarget(
            settings_path=(anchor_dir / SETTINGS_FILENAME).resolve(),
            anchor_dir=anchor_dir,
            path_mode="directory_anchor",
        )

    if path.suffix.lower() in {".yaml", ".yml"}:
        resolved = path.resolve()
        return ResolvedSettingsTarget(
            settings_path=resolved,
            anchor_dir=resolved.parent,
            path_mode="explicit_file",
        )

    anchor_dir = path.resolve()
    return ResolvedSettingsTarget(
        settings_path=(anchor_dir / SETTINGS_FILENAME).resolve(),
        anchor_dir=anchor_dir,
        path_mode="directory_anchor",
    )


def build_config_context(
    *,
    control: ControlModel,
    fallback_context: ConfigContext | None,
) -> ConfigContext:
    resolved_target = _resolve_settings_target_from_control(
        control_settings_path=control.settings.path,
        control_base_dir=control.base_dir,
        fallback_context=fallback_context,
    )
    resolved_env_prefix = resolve_env_prefix(control.settings.env_prefix)
    return ConfigContext(
        settings_path=resolved_target.settings_path,
        anchor_dir=resolved_target.anchor_dir,
        path_mode=resolved_target.path_mode,
        env_prefix=resolved_env_prefix,
        case_sensitive=control.settings.case_sensitive,
        reload_mode=control.settings.reload,
    )


def _resolve_settings_target_from_control(
    *,
    control_settings_path: str | None,
    control_base_dir: str | None,
    fallback_context: ConfigContext | None,
) -> ResolvedSettingsTarget:
    from_control = resolve_settings_target(control_settings_path)
    if from_control is not None:
        return from_control

    from_base_dir = resolve_settings_target(control_base_dir, as_directory=True)
    if from_base_dir is not None:
        return from_base_dir

    if fallback_context is not None:
        return ResolvedSettingsTarget(
            settings_path=fallback_context.settings_path,
            anchor_dir=fallback_context.anchor_dir,
            path_mode=fallback_context.path_mode,
        )

    default_target = resolve_settings_target(Path.cwd(), as_directory=True)
    assert default_target is not None
    return default_target


__all__ = [
    "build_config_context",
    "ConfigContext",
    "ResolvedSettingsTarget",
    "resolve_settings_target",
]



from __future__ import annotations

from collections.abc import Mapping
from copy import deepcopy
from dataclasses import dataclass
from typing import Any

from .types import ProjectionKind, SourceName


@dataclass(frozen=True)
class SourceEntry:
    source: SourceName
    priority: int
    kind: ProjectionKind
    include_in_control: bool
    path: tuple[str, ...]
    rev: int
    value: Any


@dataclass(frozen=True)
class EntrySource:
    source: SourceName
    priority: int
    kind: ProjectionKind
    include_in_control: bool
    rev: int
    mapping: Mapping[Any, Any]


@dataclass
class _FlattenFrame:
    prefix: tuple[str, ...]
    mapping: Mapping[Any, Any]
    items: Any


def build_entries_from_mappings(sources: list[EntrySource]) -> tuple[SourceEntry, ...]:
    rows: list[SourceEntry] = []
    for source in sources:
        flat = _flatten_mapping(source.mapping)
        for path, value in flat.items():
            rows.append(
                SourceEntry(
                    source=source.source,
                    priority=source.priority,
                    kind=source.kind,
                    include_in_control=source.include_in_control,
                    path=path,
                    rev=source.rev,
                    value=deepcopy(value),
                )
            )

    rows.sort(key=lambda row: (row.rev, row.priority, row.path, row.source))
    return tuple(rows)


def _flatten_mapping(
    mapping: Mapping[Any, Any],
    *,
    prefix: tuple[str, ...] = (),
) -> dict[tuple[str, ...], Any]:
    flat: dict[tuple[str, ...], Any] = {}

    stack: list[_FlattenFrame] = [
        _FlattenFrame(prefix=prefix, mapping=mapping, items=iter(_snapshot_mapping_items(mapping)))
    ]
    active_mapping_ids: set[int] = {id(mapping)}

    while stack:
        frame = stack[-1]
        try:
            key, value = next(frame.items)
        except StopIteration:
            active_mapping_ids.remove(id(frame.mapping))
            stack.pop()
            continue

        path = (*frame.prefix, str(key))
        if not isinstance(value, Mapping):
            flat[path] = value
            continue

        if not value:
            flat[path] = {}
            continue

        nested_id = id(value)
        if nested_id in active_mapping_ids:
            path_text = ".".join(path)
            raise ValueError(f"cyclic mapping detected at path '{path_text}'")

        active_mapping_ids.add(nested_id)
        stack.append(
            _FlattenFrame(prefix=path, mapping=value, items=iter(_snapshot_mapping_items(value)))
        )

    return flat


def _snapshot_mapping_items(mapping: Mapping[Any, Any]) -> tuple[tuple[Any, Any], ...]:
    for _ in range(2):
        try:
            return tuple(mapping.items())
        except (KeyError, RuntimeError):
            continue

    try:
        keys = tuple(mapping.keys())
    except (KeyError, RuntimeError):
        return ()

    items: list[tuple[Any, Any]] = []
    for key in keys:
        try:
            items.append((key, mapping[key]))
        except (KeyError, RuntimeError):
            continue
    return tuple(items)



from __future__ import annotations


class SettingsError(RuntimeError):
    """Base error for settings domain failures."""


class SettingsRegistrationError(SettingsError):
    """Raised when settings declarations are invalid or conflicting."""


class SettingsValidationError(SettingsError):
    """Raised when loaded raw settings fail schema validation."""


class SettingsResolveError(SettingsError):
    """Raised when a settings read cannot be resolved and no default is provided."""


__all__ = [
    "SettingsError",
    "SettingsRegistrationError",
    "SettingsValidationError",
    "SettingsResolveError",
]



from __future__ import annotations

from typing import Literal, TypeAlias

ReloadMode = Literal["off", "on_change", "always"]
ResolveAPI = Literal["settings", "map"]
SectionKind = Literal["object", "map"]
ProjectionKind = Literal["mapping", "env_like"]
SourceName: TypeAlias = str
SettingsPathMode = Literal["explicit_file", "directory_anchor"]
SourceState: TypeAlias = tuple[str, bool, int, int] | None
SourceDescriptor: TypeAlias = object
SourceToken: TypeAlias = object | None
SourceSyncMode = Literal["none", "auto", "reload", "full"]

__all__ = [
    "ReloadMode",
    "ResolveAPI",
    "SectionKind",
    "ProjectionKind",
    "SettingsPathMode",
    "SourceDescriptor",
    "SourceName",
    "SourceState",
    "SourceSyncMode",
    "SourceToken",
]



from __future__ import annotations

from dataclasses import dataclass

from pydantic import BaseModel

from .pathing import resolve_section_name, split_dotted_path
from .types import SectionKind


@dataclass(frozen=True, slots=True)
class SectionSpec:
    model: type[BaseModel]
    kind: SectionKind
    raw_path: str
    path: tuple[str, ...]

    @property
    def root(self) -> str:
        return self.path[0]

    @property
    def path_text(self) -> str:
        return ".".join(self.path)

    def path_with(self, *suffix: str) -> tuple[str, ...]:
        return (*self.path, *_normalize_suffix(suffix))

    def dotted(self, *suffix: str) -> str:
        return ".".join(self.path_with(*suffix))

    def env_key(self, *suffix: str, separator: str) -> str:
        return separator.join(part.upper() for part in self.path_with(*suffix))


def describe_section(
    model: type[BaseModel],
    *,
    kind: SectionKind,
    explicit: str | None = None,
) -> SectionSpec:
    raw_path = resolve_section_name(model, explicit)
    return SectionSpec(
        model=model,
        kind=kind,
        raw_path=raw_path,
        path=split_dotted_path(raw_path),
    )


def _normalize_suffix(suffix: tuple[str, ...]) -> tuple[str, ...]:
    normalized: list[str] = []
    for part in suffix:
        normalized.extend(split_dotted_path(part))
    return tuple(normalized)



from __future__ import annotations

from collections.abc import Mapping
from typing import Any

from .control_contract import CONTROL_SPEC, ControlModel
from .loader import load_env_overrides
from .projection import merge_nested_mapping, normalize_control_mapping


def normalize_control_snapshot(snapshot: Mapping[Any, Any]) -> dict[str, Any]:
    return _extract_control_mapping(snapshot, CONTROL_SPEC.path)


def read_control_model(snapshot: Mapping[Any, Any]) -> ControlModel:
    normalized = normalize_control_snapshot(snapshot)
    return ControlModel.model_validate(normalized)


def _extract_control_mapping(
    snapshot: Mapping[Any, Any],
    path: tuple[str, ...],
) -> dict[str, Any]:
    candidates: list[Mapping[Any, Any]] = [snapshot]

    for segment in path:
        next_candidates: list[Mapping[Any, Any]] = []
        for candidate in candidates:
            for key, value in candidate.items():
                if not isinstance(key, str):
                    continue
                if key.casefold() != segment.casefold():
                    continue
                if isinstance(value, Mapping):
                    next_candidates.append(value)
        if not next_candidates:
            return {}
        candidates = next_candidates

    merged: dict[str, Any] = {}
    for candidate in candidates:
        merge_nested_mapping(merged, normalize_control_mapping(candidate))
    return merged

def build_env_controls_snapshot() -> Mapping[Any, Any]:
    return load_env_overrides(prefix="", case_sensitive=False)


__all__ = [
    "build_env_controls_snapshot",
    "ControlModel",
    "read_control_model",
]



from __future__ import annotations

from typing import Any

from pydantic import Field

from fastapiex.settings.base import BaseSettings
from fastapiex.settings.projection import project_snapshot_for_validation


def test_snapshot_projection_projects_model_fields_case_insensitively() -> None:
    class AppSettings(BaseSettings):
        name: str

    class RootSettings(BaseSettings):
        app: AppSettings = Field(default_factory=lambda: AppSettings(name="default"))
        fastapiex: dict[str, Any] = Field(default_factory=dict)

    raw = {
        "APP": {"NAME": "demo"},
        "FastAPIEx": {"Settings": {"Reload": "always"}},
    }

    projected = project_snapshot_for_validation(
        raw,
        root_model=RootSettings,
        case_sensitive=False,
    )

    assert projected["app"]["name"] == "demo"
    assert projected["fastapiex"]["settings"]["reload"] == "always"


def test_snapshot_projection_projects_map_values_to_declared_model() -> None:
    class ServiceSettings(BaseSettings):
        host: str

    class RootSettings(BaseSettings):
        services: dict[str, ServiceSettings] = Field(default_factory=dict)
        fastapiex: dict[str, Any] = Field(default_factory=dict)

    raw = {"services": {"api": {"HOST": "127.0.0.1"}}}

    projected = project_snapshot_for_validation(
        raw,
        root_model=RootSettings,
        case_sensitive=False,
    )

    assert projected["services"]["api"]["host"] == "127.0.0.1"



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, cast

from pydantic import BaseModel, ConfigDict, Field, create_model

from .control_contract import CONTROL_SPEC
from .exceptions import SettingsRegistrationError
from .specs import SectionSpec


@dataclass(frozen=True)
class BuiltSchema:
    root_model: type[BaseModel]
    sections: tuple[SectionSpec, ...]


@dataclass
class _TreeNode:
    name: str
    decl: SectionSpec | None = None
    children: dict[str, "_TreeNode"] = field(default_factory=dict)


def build_root_settings_model(
    sections: list[SectionSpec],
    *,
    model_name: str = "FastAPIExRootSettings",
) -> BuiltSchema:
    root = _TreeNode(name="__root__")
    for section in sorted(sections, key=lambda item: item.path):
        _insert_section(root, section)

    field_defs: dict[str, tuple[Any, Any]] = {}
    for child_name, child in sorted(root.children.items()):
        if child.decl is not None and child.decl.kind == "map":
            field_defs[child_name] = _build_map_field_def(child.decl.model)
            continue

        child_model = _build_object_model(child, model_name=f"{model_name}_{child_name}")
        field_defs[child_name] = (child_model, Field(default_factory=child_model))

    if CONTROL_SPEC.root not in field_defs:
        field_defs[CONTROL_SPEC.root] = (dict[str, Any], Field(default_factory=dict))

    root_model = _create_dynamic_model(
        model_name=model_name,
        base_model=BaseModel,
        field_defs=field_defs,
    )
    return BuiltSchema(root_model=root_model, sections=tuple(sorted(sections, key=lambda item: item.path)))


def _insert_section(root: _TreeNode, section: SectionSpec) -> None:
    current = root
    for part in section.path:
        if current.decl is not None and current.decl.kind == "map":
            raise SettingsRegistrationError(
                f"map section '{current.decl.path_text}' cannot have nested section '{section.path_text}'"
            )
        child = current.children.get(part)
        if child is None:
            child = _TreeNode(name=part)
            current.children[part] = child
        current = child

    if current.children and section.kind == "map":
        raise SettingsRegistrationError(
            f"map section '{section.path_text}' conflicts with existing nested declarations"
        )

    existing = current.decl
    if existing is not None and (existing.model is not section.model or existing.kind != section.kind):
        raise SettingsRegistrationError(f"section '{section.path_text}' is declared by multiple incompatible models")
    current.decl = section


def _build_object_model(node: _TreeNode, *, model_name: str) -> type[BaseModel]:
    if node.decl is not None and node.decl.kind == "map":
        raise SettingsRegistrationError(
            f"internal error: map node '{node.decl.path_text}' must be emitted as mapping field"
        )

    base_model: type[BaseModel] = node.decl.model if node.decl is not None else BaseModel

    field_defs: dict[str, tuple[Any, Any]] = {}
    for child_name, child in sorted(node.children.items()):
        if base_model is not BaseModel and child_name in base_model.model_fields:
            owner = node.decl.path_text if node.decl is not None else model_name
            raise SettingsRegistrationError(
                f"nested declaration '{owner}.{child_name}' conflicts with existing field '{child_name}'"
            )

        if child.decl is not None and child.decl.kind == "map":
            field_defs[child_name] = _build_map_field_def(child.decl.model)
            continue

        child_model = _build_object_model(child, model_name=f"{model_name}_{child_name}")
        field_defs[child_name] = (child_model, Field(default_factory=child_model))

    return _create_dynamic_model(
        model_name=model_name,
        base_model=base_model,
        field_defs=field_defs,
    )


def _build_map_field_def(map_model: type[BaseModel]) -> tuple[Any, Any]:
    return (dict[str, map_model], Field(default_factory=dict))  # type: ignore[valid-type]


def _create_dynamic_model(
    *,
    model_name: str,
    base_model: type[BaseModel],
    field_defs: dict[str, tuple[Any, Any]],
) -> type[BaseModel]:
    return cast(
        type[BaseModel],
        create_model(  # type: ignore[call-overload]
            model_name,
            __base__=base_model,
            __config__=ConfigDict(extra="ignore"),
            **field_defs,
        ),
    )



from __future__ import annotations

from fastapiex.settings.loader import parse_dotenv_value, parse_env_value


def test_parse_env_value_parses_scalars() -> None:
    assert parse_env_value("yes") is True
    assert parse_env_value("off") is False
    assert parse_env_value("none") is None
    assert parse_env_value("1_024") == 1024
    assert parse_env_value("6.02e3") == 6020.0


def test_parse_env_value_parses_json_literals() -> None:
    assert parse_env_value("[1, 2, 3]") == [1, 2, 3]
    assert parse_env_value('{"a": 1}') == {"a": 1}


def test_parse_env_value_keeps_invalid_json_text() -> None:
    assert parse_env_value("{invalid json}") == "{invalid json}"


def test_parse_dotenv_value_strips_inline_comments_but_preserves_quoted_hash() -> None:
    assert parse_dotenv_value("abc # comment") == "abc"
    assert parse_dotenv_value('"abc # still value" # comment') == "abc # still value"



from __future__ import annotations

import logging
import os
from typing import TypeAlias

from pydantic import ConfigDict, Field, field_validator

from .constants import (
    DEFAULT_CASE_SENSITIVE,
    DEFAULT_ENV_PREFIX,
    DEFAULT_RELOAD_MODE,
    ENV_KEY_SEPARATOR,
    FALSE_TEXT_VALUES,
    RELOAD_MODE_ALWAYS_TOKENS,
    RELOAD_MODE_OFF_TOKENS,
    RELOAD_MODE_ON_CHANGE_TOKENS,
    TRUE_TEXT_VALUES,
)
from .core_settings import CoreSettings
from .specs import SectionSpec
from .types import ReloadMode

logger = logging.getLogger(__name__)


def _normalize_optional_str(raw: object | None) -> str | None:
    if raw is None:
        return None
    value = str(raw).strip()
    return value or None


def _parse_bool(raw: object | None, *, default: bool = False) -> bool:
    if raw is None:
        return default
    if isinstance(raw, bool):
        return raw
    if isinstance(raw, (int, float)):
        return bool(raw)

    value = str(raw).strip().lower()
    if value in TRUE_TEXT_VALUES:
        return True
    if value in FALSE_TEXT_VALUES:
        return False
    return default


def _parse_case_sensitive_mode(raw: object | None, *, default: bool = DEFAULT_CASE_SENSITIVE) -> bool:
    mode = _parse_bool(raw, default=default)
    if os.name == "nt" and mode:
        logger.warning("CASE_SENSITIVE=true is ignored on Windows; falling back to case-insensitive mode")
        return False
    return mode


def _parse_reload_mode(raw: object | None, *, default: ReloadMode = DEFAULT_RELOAD_MODE) -> ReloadMode:
    if isinstance(raw, bool):
        return "on_change" if raw else "off"
    if isinstance(raw, (int, float)):
        return "on_change" if raw else "off"

    raw_mode = None if raw is None else str(raw).strip().lower()
    if raw_mode is None:
        return default
    if raw_mode in RELOAD_MODE_ALWAYS_TOKENS:
        return "always"
    if raw_mode in RELOAD_MODE_ON_CHANGE_TOKENS:
        return "on_change"
    if raw_mode in RELOAD_MODE_OFF_TOKENS:
        return "off"

    logger.warning("invalid settings reload mode %r; falling back to %r", raw_mode, default)
    return default


class SettingsControls(CoreSettings):
    model_config = ConfigDict(extra="ignore")
    __section__ = "settings"

    path: str | None = None
    env_prefix: str = DEFAULT_ENV_PREFIX
    case_sensitive: bool = DEFAULT_CASE_SENSITIVE
    reload: ReloadMode = DEFAULT_RELOAD_MODE

    @field_validator("path", mode="before")
    @classmethod
    def _validate_path(cls, value: object | None) -> str | None:
        return _normalize_optional_str(value)

    @field_validator("env_prefix", mode="before")
    @classmethod
    def _validate_env_prefix(cls, value: object | None) -> str:
        if value is None:
            return DEFAULT_ENV_PREFIX
        text = str(value).strip()
        return text or DEFAULT_ENV_PREFIX

    @field_validator("case_sensitive", mode="before")
    @classmethod
    def _validate_case_sensitive(cls, value: object | None) -> bool:
        return _parse_case_sensitive_mode(value)

    @field_validator("reload", mode="before")
    @classmethod
    def _validate_reload_mode(cls, value: object | None) -> ReloadMode:
        return _parse_reload_mode(value)


class Fastapiex(CoreSettings):
    model_config = ConfigDict(extra="ignore")

    settings: SettingsControls = Field(default_factory=SettingsControls)
    base_dir: str | None = None

    @field_validator("base_dir", mode="before")
    @classmethod
    def _validate_base_dir(cls, value: object | None) -> str | None:
        return _normalize_optional_str(value)


ControlModel: TypeAlias = Fastapiex

CONTROL_SPEC: SectionSpec = ControlModel.section_spec()
CONTROL_ROOT = CONTROL_SPEC.root
CONTROL_ENV_PREFIX = f"{ControlModel.env_key(separator=ENV_KEY_SEPARATOR)}{ENV_KEY_SEPARATOR}"
SETTINGS_ENV_PREFIX_ENV_KEY = ControlModel.nested_env_key(
    SettingsControls,
    "env_prefix",
    separator=ENV_KEY_SEPARATOR,
)
SETTINGS_PATH_ENV_KEY = ControlModel.nested_env_key(
    SettingsControls,
    "path",
    separator=ENV_KEY_SEPARATOR,
)


def is_control_root(segment: str) -> bool:
    return segment.casefold() == CONTROL_ROOT.casefold()


__all__ = [
    "CONTROL_ENV_PREFIX",
    "CONTROL_ROOT",
    "CONTROL_SPEC",
    "ControlModel",
    "Fastapiex",
    "SETTINGS_PATH_ENV_KEY",
    "SETTINGS_ENV_PREFIX_ENV_KEY",
    "SettingsControls",
    "is_control_root",
]



from __future__ import annotations

from typing import Annotated, Any, get_args, get_origin

from pydantic import BaseModel

from .base import BaseSettings
from .constants import ENV_KEY_SEPARATOR
from .specs import SectionSpec, describe_section
from .types import SectionKind


class CoreSettings(BaseSettings):
    """Internal declaration base with class-derived path helpers."""

    @classmethod
    def section_spec(cls, *, kind: SectionKind = "object") -> SectionSpec:
        return describe_section(cls, kind=kind)

    @classmethod
    def section_name(cls) -> str:
        return cls.section_spec().raw_path

    @classmethod
    def section_path(cls) -> tuple[str, ...]:
        return cls.section_spec().path

    @classmethod
    def section_root(cls) -> str:
        return cls.section_spec().root

    @classmethod
    def dotted_path(cls, *suffix: str) -> str:
        return cls.section_spec().dotted(*suffix)

    @classmethod
    def env_key(cls, *suffix: str, separator: str = ENV_KEY_SEPARATOR) -> str:
        return cls.section_spec().env_key(*suffix, separator=separator)

    @classmethod
    def nested_field_name(cls, nested_model: type[BaseModel]) -> str:
        matches = [
            field_name
            for field_name, field in cls.model_fields.items()
            if _annotation_matches_model(field.annotation, nested_model)
        ]

        if not matches:
            raise ValueError(
                f"{cls.__module__}.{cls.__qualname__} does not contain a nested field for "
                f"{nested_model.__module__}.{nested_model.__qualname__}"
            )
        if len(matches) > 1:
            joined = ", ".join(sorted(matches))
            raise ValueError(
                f"{cls.__module__}.{cls.__qualname__} contains multiple nested fields for "
                f"{nested_model.__module__}.{nested_model.__qualname__}: {joined}"
            )

        return matches[0]

    @classmethod
    def nested_path(cls, nested_model: type[BaseModel], *suffix: str) -> tuple[str, ...]:
        return cls.section_spec().path_with(cls.nested_field_name(nested_model), *suffix)

    @classmethod
    def nested_dotted_path(cls, nested_model: type[BaseModel], *suffix: str) -> str:
        return ".".join(cls.nested_path(nested_model, *suffix))

    @classmethod
    def nested_env_key(
        cls,
        nested_model: type[BaseModel],
        *suffix: str,
        separator: str = ENV_KEY_SEPARATOR,
    ) -> str:
        return separator.join(part.upper() for part in cls.nested_path(nested_model, *suffix))


def _annotation_matches_model(annotation: Any, model_type: type[BaseModel]) -> bool:
    candidate = _unwrap_annotation(annotation)
    return isinstance(candidate, type) and issubclass(candidate, BaseModel) and candidate is model_type


def _unwrap_annotation(annotation: Any) -> Any:
    origin = get_origin(annotation)
    if origin is Annotated:
        args = get_args(annotation)
        if args:
            return _unwrap_annotation(args[0])

    args = get_args(annotation)
    if not args:
        return annotation

    non_none = [arg for arg in args if arg is not type(None)]
    if len(non_none) == 1 and len(non_none) != len(args):
        return _unwrap_annotation(non_none[0])

    return annotation



from __future__ import annotations

from pydantic import Field

from fastapiex.settings.constants import ENV_KEY_SEPARATOR
from fastapiex.settings.control_contract import CONTROL_SPEC, ControlModel, SETTINGS_ENV_PREFIX_ENV_KEY, SettingsControls
from fastapiex.settings.core_settings import CoreSettings


def test_control_contract_is_derived_from_control_model() -> None:
    assert CONTROL_SPEC.root == ControlModel.section_root()
    assert CONTROL_SPEC.path_text == ControlModel.section_name()
    assert SETTINGS_ENV_PREFIX_ENV_KEY == ControlModel.nested_env_key(
        SettingsControls,
        "env_prefix",
        separator=ENV_KEY_SEPARATOR,
    )


def test_core_settings_derives_nested_paths_from_class_contract() -> None:
    class RuntimeKnobs(CoreSettings):
        enabled: bool = True

    class RuntimeCore(CoreSettings):
        runtime: RuntimeKnobs = Field(default_factory=RuntimeKnobs)

    assert RuntimeCore.section_root() == "runtime_core"
    assert RuntimeCore.dotted_path() == "runtime_core"
    assert RuntimeCore.env_key(separator=ENV_KEY_SEPARATOR) == "RUNTIME_CORE"
    assert RuntimeCore.nested_field_name(RuntimeKnobs) == "runtime"
    assert RuntimeCore.nested_dotted_path(RuntimeKnobs, "enabled") == "runtime_core.runtime.enabled"
    assert RuntimeCore.nested_env_key(RuntimeKnobs, "enabled", separator=ENV_KEY_SEPARATOR) == (
        "RUNTIME_CORE__RUNTIME__ENABLED"
    )



from __future__ import annotations

from typing import Any

import pytest

from fastapiex.settings.base import BaseSettings
from fastapiex.settings.exceptions import SettingsResolveError
from fastapiex.settings.query import (
    QueryMiss,
    ResolveRequest,
    evaluate_request,
    resolve_default,
    resolve_type_target,
)
from fastapiex.settings.specs import SectionSpec
from fastapiex.settings.types import SectionKind


def _section(
    *,
    path: tuple[str, ...],
    model: type[BaseSettings],
    kind: SectionKind,
) -> SectionSpec:
    return SectionSpec(
        model=model,
        kind=kind,
        raw_path=".".join(path),
        path=path,
    )


def test_query_engine_evaluates_target_and_field() -> None:
    class AppSettings(BaseSettings):
        name: str

    class RootSettings(BaseSettings):
        app: AppSettings

    root = RootSettings(app=AppSettings(name="demo"))
    request = ResolveRequest(
        api="settings",
        target="app",
        field="name",
        default=object(),
        has_default=False,
    )

    value = evaluate_request(
        request=request,
        settings=root,
        sections=[_section(path=("app",), model=AppSettings, kind="object")],
        case_sensitive=False,
    )

    assert value == "demo"


def test_query_engine_rejects_non_mapping_for_map_api() -> None:
    class AppSettings(BaseSettings):
        name: str

    class RootSettings(BaseSettings):
        app: AppSettings

    root = RootSettings(app=AppSettings(name="demo"))
    request = ResolveRequest(
        api="map",
        target="app",
        field=None,
        default=object(),
        has_default=False,
    )

    with pytest.raises(QueryMiss, match="not a mapping"):
        evaluate_request(
            request=request,
            settings=root,
            sections=[_section(path=("app",), model=AppSettings, kind="object")],
            case_sensitive=False,
        )


def test_query_engine_requires_unique_type_target() -> None:
    class SharedMarker:
        pass

    class AppSettings(BaseSettings, SharedMarker):
        value: int

    class WorkerSettings(BaseSettings, SharedMarker):
        value: int

    sections = [
        _section(path=("app",), model=AppSettings, kind="object"),
        _section(path=("worker",), model=WorkerSettings, kind="object"),
    ]

    with pytest.raises(QueryMiss, match="matched multiple sections"):
        resolve_type_target(target_type=SharedMarker, sections=sections)


def test_query_engine_default_for_map_must_be_mapping() -> None:
    request = ResolveRequest(
        api="map",
        target=None,
        field=None,
        default=[],
        has_default=True,
    )

    with pytest.raises(SettingsResolveError, match="must be a mapping"):
        resolve_default(request)

    mapping_default: dict[str, Any] = {"ok": 1}
    request_ok = ResolveRequest(
        api="map",
        target=None,
        field=None,
        default=mapping_default,
        has_default=True,
    )
    assert resolve_default(request_ok) is mapping_default



from __future__ import annotations

import os
from pathlib import Path

from pydantic import BaseModel

from .control_contract import SETTINGS_PATH_ENV_KEY
from .manager import get_settings_manager


def init_settings(
    *,
    settings_path: str | Path | None = None,
) -> BaseModel:
    if settings_path is not None:
        os.environ[SETTINGS_PATH_ENV_KEY] = os.fspath(settings_path)
    return get_settings_manager().init()


def reload_settings(*, reason: str = "manual") -> BaseModel:
    return get_settings_manager().reload(reason=reason)



from __future__ import annotations

import os
from collections.abc import Iterator, Mapping
from pathlib import Path

import pytest

from fastapiex.settings import loader as loader_module
from fastapiex.settings.live_config import EntrySource, build_entries_from_mappings
from fastapiex.settings.loader import (
    load_dotenv_file_snapshot,
    load_env_snapshot_raw,
    load_env_overrides,
    load_yaml_file_snapshot,
    parse_env_snapshot,
    read_env_prefix_override,
    resolve_env_prefix,
)
from fastapiex.settings.projection import materialize_effective_snapshot


@pytest.fixture(autouse=True)
def _reset_env(monkeypatch: pytest.MonkeyPatch) -> None:
    controlled_prefixes = (
        "FASTAPIEX__",
        "TEST__",
        "UNITTEST__",
        "SOME_CUSTOM_PREFIX",
        "CODEX_TEST_",
    )
    for key in list(os.environ):
        if key.startswith(controlled_prefixes):
            monkeypatch.delenv(key, raising=False)


def _materialize_raw(*, path: Path, env_prefix: str, case_sensitive: bool) -> dict[str, object]:
    resolved_prefix = resolve_env_prefix(env_prefix)
    yaml_raw, _ = load_yaml_file_snapshot(path)
    dotenv_raw = parse_env_snapshot(
        load_dotenv_file_snapshot(path.parent / ".env")[0],
        prefix=resolved_prefix,
        case_sensitive=case_sensitive,
    )
    env_raw = load_env_overrides(prefix=resolved_prefix, case_sensitive=case_sensitive)
    entries = build_entries_from_mappings(
        [
            EntrySource(source="yaml", priority=1, kind="mapping", include_in_control=True, rev=1, mapping=yaml_raw),
            EntrySource(
                source="dotenv",
                priority=2,
                kind="mapping",
                include_in_control=True,
                rev=2,
                mapping=dotenv_raw,
            ),
            EntrySource(source="env", priority=3, kind="mapping", include_in_control=True, rev=3, mapping=env_raw),
        ]
    )
    return materialize_effective_snapshot(entries, env_prefix="", case_sensitive=case_sensitive)


class _FlakyEnviron(Mapping[str, str]):
    def __init__(self, data: Mapping[str, str], *, failures: int) -> None:
        self._data = dict(data)
        self._failures = failures

    def __getitem__(self, key: str) -> str:
        if self._failures > 0:
            self._failures -= 1
            raise KeyError(key)
        return self._data[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)


class _IterFlakyEnviron(Mapping[str, str]):
    def __init__(self, data: Mapping[str, str], *, iter_failures: int) -> None:
        self._data = dict(data)
        self._iter_failures = iter_failures

    def __getitem__(self, key: str) -> str:
        return self._data[key]

    def __iter__(self) -> Iterator[str]:
        if self._iter_failures > 0:
            self._iter_failures -= 1
            raise RuntimeError("unstable environ iteration")
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)


class _AlwaysBrokenEnviron(Mapping[str, str]):
    def __getitem__(self, key: str) -> str:
        raise KeyError(key)

    def __iter__(self) -> Iterator[str]:
        raise RuntimeError("environ unavailable")

    def __len__(self) -> int:
        return 0


def test_loader_stack_applies_env_dotenv_yaml_precedence_case_insensitive(
    monkeypatch,
    tmp_path: Path,
) -> None:
    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text(
        "app:\n  name: yaml\n  debug: false\n  port: 7000\n",
        encoding="utf-8",
    )
    (tmp_path / ".env").write_text(
        "TEST__APP__NAME=dotenv\nTEST__APP__DEBUG=true\n",
        encoding="utf-8",
    )

    monkeypatch.setenv("TEST__APP__PORT", "8080")

    raw = _materialize_raw(path=settings_file, env_prefix="TEST__", case_sensitive=False)

    assert raw["app"]["name"] == "dotenv"
    assert raw["app"]["debug"] is True
    assert raw["app"]["port"] == 8080


def test_loader_stack_case_sensitive_env_mapping_preserves_case(
    monkeypatch,
    tmp_path: Path,
) -> None:
    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("App:\n  Name: yaml\n", encoding="utf-8")

    monkeypatch.setenv("TEST__App__Name", "env-value")

    raw = _materialize_raw(path=settings_file, env_prefix="TEST__", case_sensitive=True)

    assert raw["App"]["Name"] == "env-value"
    assert "app" not in raw


def test_loader_stack_prefix_match_is_case_insensitive_when_case_insensitive_mode(
    monkeypatch,
    tmp_path: Path,
) -> None:
    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("app:\n  name: yaml\n", encoding="utf-8")

    monkeypatch.setenv("test__APP__NAME", "env-value")

    raw = _materialize_raw(path=settings_file, env_prefix="TEST__", case_sensitive=False)

    assert raw["app"]["name"] == "env-value"


def test_runtime_control_env_keys_are_plain_snapshot_keys(monkeypatch, tmp_path: Path) -> None:
    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("app:\n  name: yaml\n", encoding="utf-8")

    monkeypatch.setenv("FASTAPIEX__SETTINGS__CASE_SENSITIVE", "true")
    monkeypatch.setenv("FASTAPIEX__SETTINGS__RELOAD", "on_change")
    monkeypatch.setenv("FASTAPIEX__SETTINGS__PATH", "/tmp/settings.yaml")
    monkeypatch.setenv("FASTAPIEX__BASE_DIR", "/tmp")
    monkeypatch.setenv("FASTAPIEX__SETTINGS__ENV_PREFIX", "TEST__")

    raw = _materialize_raw(path=settings_file, env_prefix="", case_sensitive=False)

    assert raw["fastapiex"]["settings"]["case_sensitive"] is True
    assert raw["fastapiex"]["settings"]["reload"] == "on_change"
    assert raw["fastapiex"]["settings"]["path"] == "/tmp/settings.yaml"
    assert raw["fastapiex"]["base_dir"] == "/tmp"
    assert raw["fastapiex"]["settings"]["env_prefix"] == "TEST__"
    assert raw["app"]["name"] == "yaml"


def test_loader_stack_parses_extended_scalar_literals(monkeypatch, tmp_path: Path) -> None:
    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("app: {}\n", encoding="utf-8")

    monkeypatch.setenv("TEST__APP__TRUTHY", "yes")
    monkeypatch.setenv("TEST__APP__FALSY", "off")
    monkeypatch.setenv("TEST__APP__NONE", "none")
    monkeypatch.setenv("TEST__APP__COUNT", "1_000")
    monkeypatch.setenv("TEST__APP__RATE", "6.02e3")
    monkeypatch.setenv("TEST__APP__QUOTED", '"hello"')

    raw = _materialize_raw(path=settings_file, env_prefix="TEST__", case_sensitive=False)

    assert raw["app"]["truthy"] is True
    assert raw["app"]["falsy"] is False
    assert raw["app"]["none"] is None
    assert raw["app"]["count"] == 1000
    assert raw["app"]["rate"] == 6020.0
    assert raw["app"]["quoted"] == "hello"


def test_loader_stack_empty_prefix_reads_plain_env_key(monkeypatch, tmp_path: Path) -> None:
    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("{}", encoding="utf-8")

    monkeypatch.setenv("CODEX_TEST_ONE", "1")
    raw = _materialize_raw(path=settings_file, env_prefix="", case_sensitive=False)

    assert raw["codex_test_one"] == 1


def test_loader_stack_rejects_reserved_env_prefix_value(tmp_path: Path) -> None:
    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("{}", encoding="utf-8")

    try:
        _materialize_raw(path=settings_file, env_prefix="FASTAPIEX__", case_sensitive=False)
    except ValueError as exc:
        assert "FASTAPIEX__SETTINGS__ENV_PREFIX" in str(exc)
    else:
        raise AssertionError("expected reserved prefix rejection")


def test_loader_stack_strips_raw_prefix_without_forcing_separator(monkeypatch, tmp_path: Path) -> None:
    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("{}", encoding="utf-8")

    monkeypatch.setenv("SOME_CUSTOM_PREFIX__ONE", "1")
    monkeypatch.setenv("SOME_CUSTOM_PREFIX_TWO", "2")
    monkeypatch.setenv("SOME_CUSTOM_PREFIXTHREE", "3")

    raw_a = _materialize_raw(path=settings_file, env_prefix="SOME_CUSTOM_PREFIX__", case_sensitive=False)
    raw_b = _materialize_raw(path=settings_file, env_prefix="SOME_CUSTOM_PREFIX_", case_sensitive=False)
    raw_c = _materialize_raw(path=settings_file, env_prefix="SOME_CUSTOM_PREFIX", case_sensitive=False)

    assert raw_a["one"] == 1
    assert raw_b["two"] == 2
    assert raw_c["three"] == 3


def test_loader_stack_allows_triple_underscore_segment(monkeypatch, tmp_path: Path) -> None:
    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("{}", encoding="utf-8")

    monkeypatch.setenv("UNITTEST__FOO___BAR", "1")

    raw = _materialize_raw(path=settings_file, env_prefix="UNITTEST__", case_sensitive=False)

    assert raw["foo"]["_bar"] == 1


def test_loader_stack_ignores_quadruple_underscore_from_env(monkeypatch, tmp_path: Path) -> None:
    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("{}", encoding="utf-8")

    monkeypatch.setenv("UNITTEST__FOO____BAR", "1")

    raw = _materialize_raw(path=settings_file, env_prefix="UNITTEST__", case_sensitive=False)
    assert raw == {}


def test_loader_stack_ignores_quadruple_underscore_from_dotenv(tmp_path: Path) -> None:
    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("{}", encoding="utf-8")
    (tmp_path / ".env").write_text("UNITTEST__FOO____BAR=1\n", encoding="utf-8")

    raw = _materialize_raw(path=settings_file, env_prefix="UNITTEST__", case_sensitive=False)
    assert raw == {}


def test_prefixed_fastapiex_key_is_ignored_and_warned_from_env(
    caplog,
    monkeypatch,
    tmp_path: Path,
) -> None:
    import logging

    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("{}", encoding="utf-8")

    monkeypatch.setenv("TEST__FASTAPIEX__SETTINGS__PATH", "/should/be/ignored")

    caplog.set_level(logging.WARNING)
    raw = _materialize_raw(path=settings_file, env_prefix="TEST__", case_sensitive=False)

    assert "fastapiex" not in raw
    warning_messages = [r.getMessage() for r in caplog.records if r.levelno >= logging.WARNING]
    assert any("must not carry" in msg and "TEST__FASTAPIEX__SETTINGS__PATH" in msg for msg in warning_messages)


def test_prefixed_fastapiex_key_is_ignored_from_dotenv(tmp_path: Path) -> None:
    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("{}", encoding="utf-8")
    (tmp_path / ".env").write_text("TEST__FASTAPIEX__SETTINGS__RELOAD=always\n", encoding="utf-8")

    raw = _materialize_raw(path=settings_file, env_prefix="TEST__", case_sensitive=False)
    assert "fastapiex" not in raw


def test_real_fastapiex_key_wins_when_prefixed_variant_also_exists(
    monkeypatch,
    tmp_path: Path,
) -> None:
    settings_file = tmp_path / "settings.yaml"
    settings_file.write_text("{}", encoding="utf-8")

    monkeypatch.setenv("FASTAPIEX__SETTINGS__PATH", "/real.yaml")
    monkeypatch.setenv("TEST__FASTAPIEX__SETTINGS__PATH", "/should/be/ignored")

    raw = _materialize_raw(path=settings_file, env_prefix="TEST__", case_sensitive=False)

    assert raw["fastapiex"]["settings"]["path"] == "/real.yaml"


def test_load_env_snapshot_raw_retries_when_environ_copy_is_unstable(monkeypatch: pytest.MonkeyPatch) -> None:
    flaky = _FlakyEnviron({"APP__NAME": "demo"}, failures=1)
    monkeypatch.setattr(loader_module, "_current_environ", lambda: flaky)

    assert load_env_snapshot_raw() == {"APP__NAME": "demo"}


def test_load_env_snapshot_raw_falls_back_after_repeated_iteration_failures(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    flaky = _IterFlakyEnviron({"APP__NAME": "demo"}, iter_failures=3)
    monkeypatch.setattr(loader_module, "_current_environ", lambda: flaky)

    assert load_env_snapshot_raw() == {"APP__NAME": "demo"}


def test_read_env_prefix_override_uses_stable_snapshot_with_case_insensitive_match(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    flaky = _FlakyEnviron({"fastapiex__settings__env_prefix": " TEST__ "}, failures=1)
    monkeypatch.setattr(loader_module, "_current_environ", lambda: flaky)

    assert read_env_prefix_override() == "TEST__"


def test_read_env_prefix_override_returns_none_when_environ_snapshot_unavailable(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr(loader_module, "_current_environ", lambda: _AlwaysBrokenEnviron())
    assert read_env_prefix_override() is None


def test_load_yaml_file_snapshot_returns_empty_mapping_when_file_is_missing(tmp_path: Path) -> None:
    payload, state = load_yaml_file_snapshot(tmp_path / "missing.yaml")
    assert payload == {}
    assert state is not None
    assert state[1] is False


def test_load_dotenv_file_snapshot_returns_empty_mapping_when_file_is_missing(tmp_path: Path) -> None:
    payload, state = load_dotenv_file_snapshot(tmp_path / "missing.env")
    assert payload == {}
    assert state is not None
    assert state[1] is False
